# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import warnings
import logging

import traits.api as t
import numpy as np
from scipy import constants
import pint

import hyperspy.api as hs
from hyperspy.signal import BaseSetMetadataItems, BaseSignal
from hyperspy import utils
from hyperspy.docstrings.signal import LAZYSIGNAL_DOC
from hyperspy.ui_registry import add_gui_method, DISPLAY_DT, TOOLKIT_DT
from hyperspy.misc.utils import isiterable
from hyperspy.external.progressbar import progressbar
from hyperspy.axes import DataAxis

from .eds import EDSSpectrum, LazyEDSSpectrum
from exspy._defaults_parser import preferences
from exspy.misc import material
from exspy.misc.eds import utils as utils_eds
from exspy.misc.elements import elements as elements_db


_logger = logging.getLogger(__name__)


@add_gui_method(toolkey="exspy.microscope_parameters_EDS_TEM")
class EDSTEMParametersUI(BaseSetMetadataItems):
    beam_energy = t.Float(t.Undefined, label="Beam energy (keV)")
    real_time = t.Float(t.Undefined, label="Real time (s)")
    tilt_stage = t.Float(t.Undefined, label="Stage tilt (degree)")
    live_time = t.Float(t.Undefined, label="Live time (s)")
    probe_area = t.Float(t.Undefined, label="Beam/probe area (nm²)")
    azimuth_angle = t.Float(t.Undefined, label="Azimuth angle (degree)")
    elevation_angle = t.Float(t.Undefined, label="Elevation angle (degree)")
    energy_resolution_MnKa = t.Float(t.Undefined, label="Energy resolution MnKa (eV)")
    beam_current = t.Float(t.Undefined, label="Beam current (nA)")
    mapping = {
        "Acquisition_instrument.TEM.beam_energy": "beam_energy",
        "Acquisition_instrument.TEM.Stage.tilt_alpha": "tilt_stage",
        "Acquisition_instrument.TEM.Detector.EDS.live_time": "live_time",
        "Acquisition_instrument.TEM.Detector.EDS.azimuth_angle": "azimuth_angle",
        "Acquisition_instrument.TEM.Detector.EDS.elevation_angle": "elevation_angle",
        "Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa": "energy_resolution_MnKa",
        "Acquisition_instrument.TEM.beam_current": "beam_current",
        "Acquisition_instrument.TEM.probe_area": "probe_area",
        "Acquisition_instrument.TEM.Detector.EDS.real_time": "real_time",
    }


class EDSTEMSpectrum(EDSSpectrum):

    """Signal class for EDS spectra measured in an TEM."""

    _signal_type = "EDS_TEM"

    def __init__(self, *args, **kwards):
        super().__init__(*args, **kwards)
        # Attributes defaults
        if "Acquisition_instrument.TEM.Detector.EDS" not in self.metadata:
            if "Acquisition_instrument.SEM.Detector.EDS" in self.metadata:
                self.metadata.set_item(
                    "Acquisition_instrument.TEM",
                    self.metadata.Acquisition_instrument.SEM,
                )
                del self.metadata.Acquisition_instrument.SEM
        self._set_default_param()

    def _set_default_param(self):
        """Set to value to default (defined in preferences)"""

        mp = self.metadata
        mp.Signal.signal_type = "EDS_TEM"

        mp = self.metadata
        if "Acquisition_instrument.TEM.Stage.tilt_alpha" not in mp:
            mp.set_item(
                "Acquisition_instrument.TEM.Stage.tilt_alpha",
                preferences.EDS.eds_tilt_stage,
            )
        if "Acquisition_instrument.TEM.Detector.EDS.elevation_angle" not in mp:
            mp.set_item(
                "Acquisition_instrument.TEM.Detector.EDS.elevation_angle",
                preferences.EDS.eds_detector_elevation,
            )
        if "Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa" not in mp:
            mp.set_item(
                "Acquisition_instrument.TEM.Detector.EDS." + "energy_resolution_MnKa",
                preferences.EDS.eds_mn_ka,
            )
        if "Acquisition_instrument.TEM.Detector.EDS.azimuth_angle" not in mp:
            mp.set_item(
                "Acquisition_instrument.TEM.Detector.EDS.azimuth_angle",
                preferences.EDS.eds_detector_azimuth,
            )

    def set_microscope_parameters(
        self,
        beam_energy=None,
        live_time=None,
        tilt_stage=None,
        azimuth_angle=None,
        elevation_angle=None,
        energy_resolution_MnKa=None,
        beam_current=None,
        probe_area=None,
        real_time=None,
        display=True,
        toolkit=None,
    ):
        if set(
            [
                beam_energy,
                live_time,
                tilt_stage,
                azimuth_angle,
                elevation_angle,
                energy_resolution_MnKa,
                beam_current,
                probe_area,
                real_time,
            ]
        ) == {None}:
            tem_par = EDSTEMParametersUI(self)
            return tem_par.gui(display=display, toolkit=toolkit)
        md = self.metadata

        if beam_energy is not None:
            md.set_item("Acquisition_instrument.TEM.beam_energy ", beam_energy)
        if live_time is not None:
            md.set_item("Acquisition_instrument.TEM.Detector.EDS.live_time", live_time)
        if tilt_stage is not None:
            md.set_item("Acquisition_instrument.TEM.Stage.tilt_alpha", tilt_stage)
        if azimuth_angle is not None:
            md.set_item(
                "Acquisition_instrument.TEM.Detector.EDS.azimuth_angle", azimuth_angle
            )
        if elevation_angle is not None:
            md.set_item(
                "Acquisition_instrument.TEM.Detector.EDS.elevation_angle",
                elevation_angle,
            )
        if energy_resolution_MnKa is not None:
            md.set_item(
                "Acquisition_instrument.TEM.Detector.EDS." + "energy_resolution_MnKa",
                energy_resolution_MnKa,
            )
        if beam_current is not None:
            md.set_item("Acquisition_instrument.TEM.beam_current", beam_current)
        if probe_area is not None:
            md.set_item("Acquisition_instrument.TEM.probe_area", probe_area)
        if real_time is not None:
            md.set_item("Acquisition_instrument.TEM.Detector.EDS.real_time", real_time)

    set_microscope_parameters.__doc__ = """
        Set the microscope parameters.

        If no arguments are given, raises an interactive mode to fill
        the values.

        Parameters
        ----------
        beam_energy: float
            The energy of the electron beam in keV
        live_time : float
            In seconds
        tilt_stage : float
            In degree
        azimuth_angle : float
            In degree
        elevation_angle : float
            In degree
        energy_resolution_MnKa : float
            In eV
        beam_current: float
            In nA
        probe_area: float
            In nm²
        real_time: float
            In seconds
        {}
        {}

        Examples
        --------
        >>> s = exspy.data.EDS_TEM_FePt_nanoparticles()
        >>> print(s.metadata.Acquisition_instrument.
        >>>       TEM.Detector.EDS.energy_resolution_MnKa)
        >>> s.set_microscope_parameters(energy_resolution_MnKa=135.)
        >>> print(s.metadata.Acquisition_instrument.
        >>>       TEM.Detector.EDS.energy_resolution_MnKa)
        133.312296
        135.0

        """.format(
        DISPLAY_DT, TOOLKIT_DT
    )

    def _are_microscope_parameters_missing(self):
        """Check if the EDS parameters necessary for quantification are
        defined in metadata."""
        must_exist = (
            "Acquisition_instrument.TEM.beam_energy",
            "Acquisition_instrument.TEM.Detector.EDS.live_time",
        )

        missing_parameters = []
        for item in must_exist:
            exists = self.metadata.has_item(item)
            if exists is False:
                missing_parameters.append(item)
        if missing_parameters:
            _logger.info("Missing parameters {}".format(missing_parameters))
            return True
        else:
            return False

    def get_calibration_from(self, ref, nb_pix=1):
        """Copy the calibration and all metadata of a reference.

        Primary use: To add a calibration to ripple file from INCA
        software

        Parameters
        ----------
        ref : signal
            The reference contains the calibration in its
            metadata
        nb_pix : int
            The live time (real time corrected from the "dead time")
            is divided by the number of pixel (spectrums), giving an
            average live time.

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.

        Examples
        --------
        >>> ref = exspy.data.EDS_TEM_FePt_nanoparticles()
        >>> s = exspy.data.EDS_TEM_FePt_nanoparticles(ref.data)
        >>> print(s.axes_manager[0].scale)
        >>> s.get_calibration_from(ref)
        >>> print(s.axes_manager[0].scale)
        1.0
        0.020028

        """

        self._original_metadata = ref.original_metadata.deepcopy()
        # Setup the axes_manager
        ax_m = self.axes_manager.signal_axes[0]
        ax_ref = ref.axes_manager.signal_axes[0]
        for _axis in [ax_m, ax_ref]:
            if not _axis.is_uniform:
                raise NotImplementedError(
                    "The function is not implemented for non-uniform axes."
                )
        ax_m.scale = ax_ref.scale
        ax_m.units = ax_ref.units
        ax_m.offset = ax_ref.offset

        # Setup metadata
        if "Acquisition_instrument.TEM" in ref.metadata:
            mp_ref = ref.metadata.Acquisition_instrument.TEM
        elif "Acquisition_instrument.SEM" in ref.metadata:
            mp_ref = ref.metadata.Acquisition_instrument.SEM
        else:
            raise ValueError(
                "The reference has no metadata "
                "'Acquisition_instrument.TEM '"
                "or 'metadata.Acquisition_instrument.SEM'."
            )

        mp = self.metadata
        mp.Acquisition_instrument.TEM = mp_ref.deepcopy()
        if mp_ref.has_item("Detector.EDS.live_time"):
            mp.Acquisition_instrument.TEM.Detector.EDS.live_time = (
                mp_ref.Detector.EDS.live_time / nb_pix
            )

    def quantification(
        self,
        intensities,
        method,
        factors,
        composition_units="atomic",
        absorption_correction=False,
        take_off_angle="auto",
        thickness="auto",
        convergence_criterion=0.5,
        navigation_mask=1.0,
        closing=True,
        plot_result=False,
        probe_area="auto",
        max_iterations=30,
        show_progressbar=None,
        **kwargs,
    ):
        """
        Absorption corrected quantification using Cliff-Lorimer, the zeta-factor
        method, or ionization cross sections. The function iterates through
        quantification function until two successive interations don't change
        the final composition by a defined percentage critera (0.5% by default).

        Parameters
        ----------
        intensities: list of signal
            the intensitiy for each X-ray lines.
        method: {'CL', 'zeta', 'cross_section'}
            Set the quantification method: Cliff-Lorimer, zeta-factor, or
            ionization cross sections.
        factors: list of float
            The list of kfactors, zeta-factors or cross sections in same order
            as intensities. Note that intensities provided by Hyperspy are
            sorted by the alphabetical order of the X-ray lines.
            eg. factors =[0.982, 1.32, 1.60] for ['Al_Ka', 'Cr_Ka', 'Ni_Ka'].
        composition_units: {'atomic', 'weight'}
            The quantification returns the composition in 'atomic' percent by
            default, but can also return weight percent if specified.
        absorption_correction: bool
            Specify whether or not an absorption correction should be applied.
            'False' by default so absorption will not be applied unless
            specfied.
        take_off_angle : {'auto'}
            The angle between the sample surface and the vector along which
            X-rays travel to reach the centre of the detector.
        thickness: {'auto'}
            thickness in nm (can be a single value or
            have the same navigation dimension as the signal).
            NB: Must be specified for 'CL' method. For 'zeta' or 'cross_section'
            methods, first quantification step provides a mass_thickness
            internally during quantification.
        convergence_criterion: The convergence criterium defined as the percentage
            difference between 2 successive iterations. 0.5% by default.
        navigation_mask : None or float or signal
            The navigation locations marked as True are not used in the
            quantification. If float is given the vacuum_mask method is used to
            generate a mask with the float value as threhsold.
            Else provides a signal with the navigation shape. Only for the
            'Cliff-Lorimer' method.
        closing: bool
            If true, applied a morphologic closing to the mask obtained by
            vacuum_mask.
        plot_result : bool
            If True, plot the calculated composition. If the current
            object is a single spectrum it prints the result instead.
        probe_area = {'auto'}
            This allows the user to specify the probe_area for interaction with
            the sample needed specifically for the cross_section method of
            quantification. When left as 'auto' the pixel area is used,
            calculated from the navigation axes information.
        max_iterations : int
            An upper limit to the number of calculations for absorption correction.
        kwargs
            The extra keyword arguments are passed to plot.

        Returns
        -------
        A list of quantified elemental maps (signal) giving the composition of
        the sample in weight or atomic percent with absorption correciton taken
        into account based on the sample thickness estimate provided.

        If the method is 'zeta' this function also returns the mass thickness
        profile for the data.

        If the method is 'cross_section' this function also returns the atom
        counts for each element.

        Examples
        --------
        >>> s = exspy.data.EDS_TEM_FePt_nanoparticles()
        >>> s.add_lines()
        >>> kfactors = [1.450226, 5.075602] #For Fe Ka and Pt La
        >>> bw = s.estimate_background_windows(line_width=[5.0, 2.0])
        >>> s.plot(background_windows=bw)
        >>> intensities = s.get_lines_intensity(background_windows=bw)
        >>> res = s.quantification(intensities, kfactors, plot_result=True,
        >>>                        composition_units='atomic')
        Fe (Fe_Ka): Composition = 15.41 atomic percent
        Pt (Pt_La): Composition = 84.59 atomic percent

        See also
        --------
        vacuum_mask
        """
        if not isinstance(intensities, (list, tuple)) or not isinstance(
            intensities[0], BaseSignal
        ):
            raise ValueError("The parameter `intensities` must be a list of signals.")
        elif len(intensities) <= 1:
            raise ValueError("Several X-ray line intensities are required.")

        if isinstance(navigation_mask, float):
            if self.axes_manager.navigation_dimension > 0:
                navigation_mask = self.vacuum_mask(navigation_mask, closing)
            else:
                navigation_mask = None

        xray_lines = [
            intensity.metadata.Sample.xray_lines[0] for intensity in intensities
        ]
        it = 0
        if absorption_correction:
            if show_progressbar is None:  # pragma: no cover
                show_progressbar = hs.preferences.General.show_progressbar
            if show_progressbar:
                pbar = progressbar(total=None, desc="Absorption correction calculation")

        composition = utils.stack(intensities, lazy=False, show_progressbar=False)

        if take_off_angle == "auto":
            toa = self.get_take_off_angle()
        else:
            toa = take_off_angle

        # determining illumination area for cross sections quantification.
        if method == "cross_section":
            if probe_area == "auto":
                parameters = self.metadata.Acquisition_instrument.TEM
                if probe_area in parameters:
                    probe_area = parameters.TEM.probe_area
                else:
                    probe_area = self.get_probe_area(
                        navigation_axes=self.axes_manager.navigation_axes
                    )

        int_stack = utils.stack(intensities, lazy=False, show_progressbar=False)
        comp_old = np.zeros_like(int_stack.data)

        abs_corr_factor = None  # initial

        if method == "CL":
            quantification_method = utils_eds.quantification_cliff_lorimer
            kwargs = {
                "intensities": int_stack.data,
                "kfactors": factors,
                "absorption_correction": abs_corr_factor,
                "mask": navigation_mask,
            }

        elif method == "zeta":
            quantification_method = utils_eds.quantification_zeta_factor
            kwargs = {
                "intensities": int_stack.data,
                "zfactors": factors,
                "dose": self._get_dose(method),
                "absorption_correction": abs_corr_factor,
            }

        elif method == "cross_section":
            quantification_method = utils_eds.quantification_cross_section
            kwargs = {
                "intensities": int_stack.data,
                "cross_sections": factors,
                "dose": self._get_dose(method, **kwargs),
                "absorption_correction": abs_corr_factor,
            }

        else:
            raise ValueError(
                "Please specify method for quantification, "
                'as "CL", "zeta" or "cross_section".'
            )

        while True:
            results = quantification_method(**kwargs)

            if method == "CL":
                composition.data = results * 100.0
                if absorption_correction:
                    if thickness is not None:
                        mass_thickness = intensities[0].deepcopy()
                        mass_thickness.data = self.CL_get_mass_thickness(
                            composition.split(), thickness
                        )
                        mass_thickness.metadata.General.title = "Mass thickness"
                    else:
                        raise ValueError(
                            "Thickness is required for absorption correction "
                            "with k-factor method. Results will contain no "
                            "correction for absorption."
                        )

            elif method == "zeta":
                composition.data = results[0] * 100
                mass_thickness = intensities[0].deepcopy()
                mass_thickness.data = results[1]

            else:
                composition.data = results[0] * 100.0
                number_of_atoms = composition._deepcopy_with_new_data(results[1])

            if method == "cross_section":
                if absorption_correction:
                    abs_corr_factor = utils_eds.get_abs_corr_cross_section(
                        composition.split(), number_of_atoms.split(), toa, probe_area
                    )
                    kwargs["absorption_correction"] = abs_corr_factor
            else:
                if absorption_correction:
                    abs_corr_factor = utils_eds.get_abs_corr_zeta(
                        composition.split(), mass_thickness, toa
                    )
                    kwargs["absorption_correction"] = abs_corr_factor

            res_max = np.max(composition.data - comp_old)
            comp_old = composition.data

            if absorption_correction and show_progressbar:
                pbar.update(1)
            it += 1
            if not absorption_correction or abs(res_max) < convergence_criterion:
                break
            elif it >= max_iterations:
                raise Exception(
                    "Absorption correction failed as solution "
                    f"did not converge after {max_iterations} "
                    "iterations"
                )

        if method == "cross_section":
            number_of_atoms = composition._deepcopy_with_new_data(results[1])
            number_of_atoms = number_of_atoms.split()
            composition = composition.split()
        else:
            composition = composition.split()

        # convert ouput units to selection as required.
        if composition_units == "atomic":
            if method != "cross_section":
                composition = material.weight_to_atomic(composition)
        else:
            if method == "cross_section":
                composition = material.atomic_to_weight(composition)

        # Label each of the elemental maps in the image stacks for composition.
        for i, xray_line in enumerate(xray_lines):
            element, line = utils_eds._get_element_and_line(xray_line)
            composition[i].metadata.General.title = (
                composition_units + " percent of " + element
            )
            composition[i].metadata.set_item("Sample.elements", ([element]))
            composition[i].metadata.set_item("Sample.xray_lines", ([xray_line]))
            if plot_result and composition[i].axes_manager.navigation_size == 1:
                c = float(composition[i].data)
                print(f"{element} ({xray_line}): Composition = {c:.2f} percent")
        # For the cross section method this is repeated for the number of atom maps
        if method == "cross_section":
            for i, xray_line in enumerate(xray_lines):
                element, line = utils_eds._get_element_and_line(xray_line)
                number_of_atoms[i].metadata.General.title = "atom counts of " + element
                number_of_atoms[i].metadata.set_item("Sample.elements", ([element]))
                number_of_atoms[i].metadata.set_item("Sample.xray_lines", ([xray_line]))
        if plot_result and composition[i].axes_manager.navigation_size != 1:
            utils.plot.plot_signals(composition, **kwargs)

        if absorption_correction:
            _logger.info(f"Convergence reached after {it} interations.")

        if method == "zeta":
            mass_thickness.metadata.General.title = "Mass thickness"
            self.metadata.set_item("Sample.mass_thickness", mass_thickness)
            return composition, mass_thickness
        elif method == "cross_section":
            return composition, number_of_atoms
        elif method == "CL":
            if absorption_correction:
                mass_thickness.metadata.General.title = "Mass thickness"
                return composition, mass_thickness
            else:
                return composition
        else:
            raise ValueError(
                "Please specify method for quantification, as "
                '"CL", "zeta" or "cross_section"'
            )

    def vacuum_mask(self, threshold=1.0, closing=True, opening=False):
        """
        Generate mask of the vacuum region

        Parameters
        ----------
        threshold: float
            For a given pixel, maximum value in the energy axis below which the
            pixel is considered as vacuum.
        closing: bool
            If true, applied a morphologic closing to the mask
        opnening: bool
            If true, applied a morphologic opening to the mask

        Returns
        -------
        mask: signal
            The mask of the region

        Examples
        --------
        >>> # Simulate a spectrum image with vacuum region
        >>> s = exspy.data.EDS_TEM_FePt_nanoparticles()
        >>> s_vac = hs.signals.BaseSignal(
                np.ones_like(s.data, dtype=float))*0.005
        >>> s_vac.add_poissonian_noise()
        >>> si = hs.stack([s]*3 + [s_vac])
        >>> si.vacuum_mask().data
        array([False, False, False,  True], dtype=bool)
        """
        if self.axes_manager.navigation_dimension == 0:
            raise RuntimeError(
                "Navigation dimenstion must be higher than 0 "
                "to estimate a vacuum mask."
            )
        from scipy.ndimage import binary_dilation, binary_erosion

        mask = self.max(-1) <= threshold
        if closing:
            mask.data = binary_dilation(mask.data, border_value=0)
            mask.data = binary_erosion(mask.data, border_value=1)
        if opening:
            mask.data = binary_erosion(mask.data, border_value=1)
            mask.data = binary_dilation(mask.data, border_value=0)
        return mask

    def decomposition(
        self,
        normalize_poissonian_noise=True,
        navigation_mask=1.0,
        closing=True,
        *args,
        **kwargs,
    ):
        """Apply a decomposition to a dataset with a choice of algorithms.

        The results are stored in ``self.learning_results``.

        Read more in the :ref:`User Guide <mva.decomposition>`.

        Parameters
        ----------
        normalize_poissonian_noise : bool, default True
            If True, scale the signal to normalize Poissonian noise using
            the approach described in [*]_.
        navigation_mask : None or float or boolean numpy array, default 1.0
            The navigation locations marked as True are not used in the
            decomposition. If float is given the vacuum_mask method is used to
            generate a mask with the float value as threshold.
        closing: bool, default True
            If true, applied a morphologic closing to the mask obtained by
            vacuum_mask.
        algorithm : {"SVD", "MLPCA", "sklearn_pca", "NMF", "sparse_pca", "mini_batch_sparse_pca", "RPCA", "ORPCA", "ORNMF", custom object}, default "SVD"
            The decomposition algorithm to use. If algorithm is an object,
            it must implement a ``fit_transform()`` method or ``fit()`` and
            ``transform()`` methods, in the same manner as a scikit-learn estimator.
        output_dimension : None or int
            Number of components to keep/calculate.
            Default is None, i.e. ``min(data.shape)``.
        centre : {None, "navigation", "signal"}, default None
            * If None, the data is not centered prior to decomposition.
            * If "navigation", the data is centered along the navigation axis.
              Only used by the "SVD" algorithm.
            * If "signal", the data is centered along the signal axis.
              Only used by the "SVD" algorithm.
        auto_transpose : bool, default True
            If True, automatically transposes the data to boost performance.
            Only used by the "SVD" algorithm.
        signal_mask : boolean numpy array
            The signal locations marked as True are not used in the
            decomposition.
        var_array : numpy array
            Array of variance for the maximum likelihood PCA algorithm.
            Only used by the "MLPCA" algorithm.
        var_func : None or function or numpy array, default None
            * If None, ignored
            * If function, applies the function to the data to obtain ``var_array``.
              Only used by the "MLPCA" algorithm.
            * If numpy array, creates ``var_array`` by applying a polynomial function
              defined by the array of coefficients to the data. Only used by
              the "MLPCA" algorithm.
        reproject : {None, "signal", "navigation", "both"}, default None
            If not None, the results of the decomposition will be projected in
            the selected masked area.
        return_info: bool, default False
            The result of the decomposition is stored internally. However,
            some algorithms generate some extra information that is not
            stored. If True, return any extra information if available.
            In the case of sklearn.decomposition objects, this includes the
            sklearn Estimator object.
        print_info : bool, default True
            If True, print information about the decomposition being performed.
            In the case of sklearn.decomposition objects, this includes the
            values of all arguments of the chosen sklearn algorithm.
        svd_solver : {"auto", "full", "arpack", "randomized"}, default "auto"
            If auto:
                The solver is selected by a default policy based on `data.shape` and
                `output_dimension`: if the input data is larger than 500x500 and the
                number of components to extract is lower than 80% of the smallest
                dimension of the data, then the more efficient "randomized"
                method is enabled. Otherwise the exact full SVD is computed and
                optionally truncated afterwards.
            If full:
                run exact SVD, calling the standard LAPACK solver via
                :py:func:`scipy.linalg.svd`, and select the components by postprocessing
            If arpack:
                use truncated SVD, calling ARPACK solver via
                :py:func:`scipy.sparse.linalg.svds`. It requires strictly
                `0 < output_dimension < min(data.shape)`
            If randomized:
                use truncated SVD, calling :py:func:`sklearn.utils.extmath.randomized_svd`
                to estimate a limited number of components
        copy : bool, default True
            * If True, stores a copy of the data before any pre-treatments
              such as normalization in ``s._data_before_treatments``. The original
              data can then be restored by calling ``s.undo_treatments()``.
            * If False, no copy is made. This can be beneficial for memory
              usage, but care must be taken since data will be overwritten.
        **kwargs : extra keyword arguments
            Any keyword arguments are passed to the decomposition algorithm.


        Examples
        --------
        >>> s = exspy.data.EDS_TEM_FePt_nanoparticles()
        >>> si = hs.stack([s]*3)
        >>> si.change_dtype(float)
        >>> si.decomposition()

        See also
        --------
        vacuum_mask

        References
        ----------
        .. [*] M. Keenan and P. Kotula, "Accounting for Poisson noise
           in the multivariate analysis of ToF-SIMS spectrum images", Surf.
           Interface Anal 36(3) (2004): 203-212.
        """
        if isinstance(navigation_mask, float):
            navigation_mask = self.vacuum_mask(navigation_mask, closing)
        super().decomposition(
            normalize_poissonian_noise=normalize_poissonian_noise,
            navigation_mask=navigation_mask,
            *args,
            **kwargs,
        )
        self.learning_results.loadings = np.nan_to_num(self.learning_results.loadings)

    def create_model(self, auto_background=True, auto_add_lines=True, *args, **kwargs):
        """Create a model for the current TEM EDS data.

        Parameters
        ----------
        auto_background : bool, default True
            If True, adds automatically a polynomial order 6 to the model,
            using the edsmodel.add_polynomial_background method.
        auto_add_lines : bool, default True
            If True, automatically add Gaussians for all X-rays generated in
            the energy range by an element using the edsmodel.add_family_lines
            method.
        dictionary : {None, dict}, optional
            A dictionary to be used to recreate a model. Usually generated
            using :meth:`hyperspy.model.as_dictionary`

        Returns
        -------
        model : `EDSTEMModel` instance.

        """
        from exspy.models.edstemmodel import EDSTEMModel

        model = EDSTEMModel(
            self,
            auto_background=auto_background,
            auto_add_lines=auto_add_lines,
            *args,
            **kwargs,
        )
        return model

    def get_probe_area(self, navigation_axes=None):
        """
        Calculates a pixel area which can be approximated to probe area,
        when the beam is larger than or equal to pixel size.
        The probe area can be calculated only when the number of navigation
        dimension are less than 2 and all the units have the dimensions of
        length.

        Parameters
        ----------
        navigation_axes : DataAxis, string or integer (or list of)
            Navigation axes corresponding to the probe area. If string or
            integer, the provided value is used to index the ``axes_manager``.

        Returns
        -------
        probe area in nm².

        Examples
        --------
        >>> s = exspy.data.EDS_TEM_FePt_nanoparticles()
        >>> si = hs.stack([s]*3)
        >>> si.axes_manager.navigation_axes[0].scale = 0.01
        >>> si.axes_manager.navigation_axes[0].units = 'μm'
        >>> si.get_probe_area()
        100.0

        """
        if navigation_axes is None:
            navigation_axes = self.axes_manager.navigation_axes
        elif not isiterable(navigation_axes):
            navigation_axes = [navigation_axes]
        if len(navigation_axes) == 0:
            raise ValueError(
                "The navigation dimension is zero, the probe "
                "area can not be calculated automatically."
            )
        elif len(navigation_axes) > 2:
            raise ValueError(
                "The navigation axes corresponding to the probe "
                "are ambiguous and the probe area can not be "
                "calculated automatically."
            )
        scales = []

        for axis in navigation_axes:
            try:
                if not isinstance(navigation_axes, DataAxis):
                    axis = self.axes_manager[axis]
                scales.append(axis.convert_to_units("nm", inplace=False)[0])
            except pint.DimensionalityError:
                raise ValueError(
                    f"The unit of the axis {axis} has not the " "dimension of length."
                )

        if len(scales) == 1:
            probe_area = scales[0] ** 2
        else:
            probe_area = scales[0] * scales[1]

        if probe_area == 1:
            warnings.warn(
                "Please note that the probe area has been "
                "calculated to be 1 nm², meaning that it is highly "
                "likley that the scale of the navigation axes have not "
                "been set correctly. Please read the user "
                "guide for how to set this."
            )
        return probe_area

    def _get_dose(
        self, method, beam_current="auto", live_time="auto", probe_area="auto"
    ):
        """
        Calculates the total electron dose for the zeta-factor or cross section
        methods of quantification.

        Input given by i*t*N, i the current, t the
        acquisition time, and N the number of electron by unit electric charge.

        Parameters
        ----------
        method : 'zeta' or 'cross_section'
            If 'zeta', the dose is given by i*t*N
            If 'cross section', the dose is given by i*t*N/A
            where i is the beam current, t is the acquistion time,
            N is the number of electrons per unit charge (1/e) and
            A is the illuminated beam area or pixel area.
        beam_current: float
            Probe current in nA
        live_time: float
            Acquisiton time in s, compensated for the dead time of the detector.
        probe_area: float or 'auto'
            The illumination area of the electron beam in nm².
            If 'auto' the value is extracted from the scale axes_manager.
            Therefore we assume the probe is oversampling such that
            the illumination area can be approximated to the pixel area of the
            spectrum image.

        Returns
        --------
        Dose in electrons (zeta factor) or electrons per nm² (cross_section)

        See also
        --------
        set_microscope_parameters
        """

        parameters = self.metadata.Acquisition_instrument.TEM

        if beam_current == "auto":
            beam_current = parameters.get_item("beam_current")
            if beam_current is None:
                raise Exception(
                    "Electron dose could not be calculated as the "
                    "beam current is not set. It can set using "
                    "`set_microscope_parameters()`."
                )

        if live_time == "auto":
            live_time = parameters.get_item("Detector.EDS.live_time")
            if live_time is None:
                raise Exception(
                    "Electron dose could not be calculated as "
                    "live time is not set. It can set using "
                    "`set_microscope_parameters()`."
                )

        if method == "cross_section":
            if probe_area == "auto":
                probe_area = parameters.get_item("probe_area")
                if probe_area is None:
                    probe_area = self.get_probe_area(
                        navigation_axes=self.axes_manager.navigation_axes
                    )
            return (live_time * beam_current * 1e-9) / (constants.e * probe_area)
            # 1e-9 is included here because the beam_current is in nA.
        elif method == "zeta":
            return live_time * beam_current * 1e-9 / constants.e
        else:
            raise Exception("Method need to be 'zeta' or 'cross_section'.")

    @staticmethod
    def CL_get_mass_thickness(weight_percent, thickness):
        """
        Creates a array of mass_thickness based on a known material composition
        and measured thickness. Required for absorption correction calcultions
        using the Cliff Lorimer method.

        Parameters
        ----------
        weight_percent : :py:class:`~hyperspy.signal.BaseSignal` (or subclass)
            Stack of compositions as determined from an initial k_factor
            quantification.
        thickness : float or :py:class:`numpy.ndarray`
            Either a float value for thickness in nm or an array equal to the
            size of the EDX map with thickness at each position of the sample.

        Returns
        -------
        mass_thickness : :py:class:`numpy.ndarray`
            Mass thickness in kg/m².
        """
        if isinstance(thickness, (float, int)):
            thickness_map = np.ones_like(weight_percent[0]) * thickness
        else:
            thickness_map = thickness

        elements = [
            intensity.metadata.Sample.elements[0] for intensity in weight_percent
        ]
        mass_thickness = np.zeros_like(weight_percent[0])
        densities = np.array(
            [
                elements_db[element]["Physical_properties"]["density (g/cm^3)"]
                for element in elements
            ]
        )
        for density, element_composition in zip(densities, weight_percent):
            # convert composition from % to fraction: factor of 1E-2
            # convert thickness from nm to m: factor of 1E-9
            # convert density from g/cm3 to kg/m2: factor of 1E3
            elemental_mt = element_composition * thickness_map * density * 1e-8
            mass_thickness += elemental_mt
        return mass_thickness


class LazyEDSTEMSpectrum(EDSTEMSpectrum, LazyEDSSpectrum):

    """Lazy signal class for EDS spectra measured in an TEM."""

    __doc__ += LAZYSIGNAL_DOC.replace("__BASECLASS__", "EDSTEMSpectrum")



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


"""Common docstring snippets for model.

"""
from exspy.misc.eels.gosh_gos import _GOSH_DOI

GOS_PARAMETER = """GOS : 'hydrogenic', 'gosh', 'Hartree-Slater'.
            The GOS to use. Default is ``'gosh'``.

        gos_file_path : str, None
            Only with GOS='gosh'. Specify the file path of the gosh file
            to use. If None, use the file from doi:{}""".format(
    _GOSH_DOI
)

EELSMODEL_PARAMETERS = """ll : None or EELSSpectrum
            If an EELSSpectrum is provided, it will be assumed that it is
            a low-loss EELS spectrum, and it will be used to simulate the
            effect of multiple scattering by convolving it with the EELS
            spectrum.
        auto_background : bool
            If True, and if spectrum is an EELS instance adds automatically
            a powerlaw to the model and estimate the parameters by the
            two-area method.
        auto_add_edges : bool
            If True, and if spectrum is an EELS instance, it will automatically
            add the ionization edges as defined in the
            :class:`~.api.signals.EELSSpectrum` instance. Adding a new element to
            the spectrum using the :meth:`~.api.signals.EELSSpectrum.add_elements`
            method automatically add the corresponding ionisation edges to the model.
        {}
        dictionary : None or dict
            A dictionary to be used to recreate a model. Usually generated using
            :meth:`~.model.BaseModel.as_dictionary`""".format(
    GOS_PARAMETER
)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy.components1d import Lorentzian

from exspy.components import VolumePlasmonDrude
from exspy.misc.eels.tools import eels_constant
from exspy.signals import EELSSpectrum


class Test2D:
    def setup_method(self, method):
        """To test the kramers_kronig_analysis we will generate 3
        EELSSpectrum instances. First a model energy loss function(ELF),
        in our case following the Drude bulk plasmon peak. Second, we
        simulate the inelastic scattering to generate a model scattering
        distribution (SPC). Finally, we use a lorentzian peak with
        integral equal to 1 to simulate a ZLP.

        """

        # Parameters
        i0 = 1.0
        t = hs.signals.BaseSignal(np.arange(10, 70, 10).reshape((2, 3)))
        t = t.transpose(signal_axes=0)
        scale = 0.02

        # Create an 3x2x2048 spectrum with Drude plasmon
        s = EELSSpectrum(np.zeros((2, 3, 2 * 2048)))
        s.set_microscope_parameters(
            beam_energy=300.0, convergence_angle=5, collection_angle=10.0
        )
        s.axes_manager.signal_axes[0].scale = scale
        k = eels_constant(s, i0, t)

        vpm = VolumePlasmonDrude()
        m = s.create_model(auto_background=False)
        m.append(vpm)
        vpm.intensity.map["values"][:] = 1
        vpm.plasmon_energy.map["values"] = np.array(
            [[8.0, 18.4, 15.8], [16.6, 4.3, 3.7]]
        )
        vpm.fwhm.map["values"] = np.array([[2.3, 4.8, 0.53], [3.7, 0.3, 0.3]])
        vpm.intensity.map["is_set"][:] = True
        vpm.plasmon_energy.map["is_set"][:] = True
        vpm.fwhm.map["is_set"][:] = True
        s.data = (m.as_signal() * k).data

        # Create ZLP
        z = s.deepcopy()
        z.axes_manager.signal_axes[0].scale = scale
        z.axes_manager.signal_axes[0].offset = -10
        zlp = Lorentzian()
        zlp.A.value = i0
        zlp.gamma.value = 0.2
        zlp.centre.value = 0.0
        z.data[:] = zlp.function(z.axes_manager[-1].axis).reshape((1, 1, -1))
        z.data *= scale
        self.s = s
        self.thickness = t
        self.k = k
        self.zlp = z

    def test_df_given_n(self):
        """The kramers kronig analysis method applied to the signal we
        have just designed above will return the CDF for the Drude bulk
        plasmon. Hopefully, we recover the signal by inverting the CDF.

        """
        # i use n=1000 to simulate a metal (enormous n)
        cdf = self.s.kramers_kronig_analysis(zlp=self.zlp, iterations=1, n=1000.0)
        s = cdf.get_electron_energy_loss_spectrum(self.zlp, self.thickness)
        np.testing.assert_allclose(s.data, self.s.data[..., 1:], rtol=0.01)

    def test_df_given_thickness(self):
        """The kramers kronig analysis method applied to the signal we
        have just designed above will return the CDF for the Drude bulk
        plasmon. Hopefully, we recover the signal by inverting the CDF.

        """
        cdf = self.s.kramers_kronig_analysis(
            zlp=self.zlp, iterations=1, t=self.thickness
        )
        s = cdf.get_electron_energy_loss_spectrum(self.zlp, self.thickness)
        np.testing.assert_allclose(s.data, self.s.data[..., 1:], rtol=0.01)

    def test_bethe_sum_rule(self):
        df = self.s.kramers_kronig_analysis(zlp=self.zlp, iterations=1, n=1000.0)
        neff1, neff2 = df.get_number_of_effective_electrons(nat=50e27, cumulative=False)
        np.testing.assert_allclose(
            neff1.data,
            np.array(
                [
                    [0.91187657, 4.72490711, 3.60594653],
                    [3.88077047, 0.26759741, 0.19813647],
                ]
            ),
            rtol=1e-6,
        )
        np.testing.assert_allclose(
            neff2.data,
            np.array(
                [
                    [0.91299039, 4.37469112, 3.41580094],
                    [3.64866394, 0.15693674, 0.11146413],
                ]
            ),
            rtol=1e-6,
        )

    def test_thickness_estimation(self):
        """Kramers kronig analysis gives a rough estimation of sample
        thickness. As we have predefined sample thickness for our
        scattering distribution, we can use it for testing putposes.

        """
        cdf, output = self.s.kramers_kronig_analysis(
            zlp=self.zlp, iterations=1, n=1000.0, full_output=True
        )
        np.testing.assert_allclose(
            self.thickness.data, output["thickness"].data, rtol=0.01
        )

    def test_thicness_input_array(self):
        with pytest.raises(ValueError):
            self.s.kramers_kronig_analysis(
                zlp=self.zlp, iterations=1, t=self.thickness.data
            )

    def test_single_spectrum_dielectric(self):
        s_in = self.s.inav[0, 0]
        z = self.zlp.inav[0, 0]
        t = self.thickness.data[0, 0]
        cdf = s_in.kramers_kronig_analysis(zlp=z, iterations=1, n=1000.0)
        s_out = cdf.get_electron_energy_loss_spectrum(z, t)
        np.testing.assert_allclose(s_out.data, s_in.data[1:], rtol=0.01)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import contextlib
import io
from unittest import mock

import numpy as np
import pooch
import pytest

import hyperspy.api as hs
from exspy.misc.elements import elements_db as elements
from hyperspy.decorators import lazifyTestClass
from exspy.misc.eels.gosh_gos import _GOSH_URL, _GOSH_KNOWN_HASH
from exspy.signals import EELSSpectrum


# Dask does not always work nicely with np.errstate,
# see: https://github.com/dask/dask/issues/3245, so
# filter out divide-by-zero warnings that only appear
# when the test is lazy. When the test is not lazy,
# internal use of np.errstate means the warnings never
# appear in the first place.
@pytest.mark.filterwarnings(
    "ignore:invalid value encountered in subtract:RuntimeWarning"
)
@pytest.mark.filterwarnings("ignore:divide by zero encountered in log:RuntimeWarning")
@lazifyTestClass
class TestCreateEELSModel:
    def setup_method(self, method):
        s = EELSSpectrum(np.zeros(200))
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        s.add_elements(("B", "C"))
        self.s = s

    def test_create_eelsmodel(self):
        from exspy.models.eelsmodel import EELSModel

        assert isinstance(self.s.create_model(), EELSModel)

    def test_create_eelsmodel_no_md(self):
        s = self.s
        del s.metadata.Acquisition_instrument
        with pytest.raises(ValueError):
            s.create_model()

    def test_auto_add_edges_true(self):
        m = self.s.create_model(auto_add_edges=True)
        cnames = [component.name for component in m]
        assert "B_K" in cnames and "C_K" in cnames

    def test_gos_hydrogenic(self):
        m = self.s.create_model(auto_add_edges=True, GOS="hydrogenic")
        assert m["B_K"].GOS._name == "hydrogenic"
        m.fit()

    def test_gos_gosh(self):
        m = self.s.create_model(auto_add_edges=True, GOS="gosh")
        assert m["B_K"].GOS._name == "gosh"
        m.fit()

        with pytest.raises(ValueError):
            self.s.create_model(auto_add_edges=True, GOS="not_a_GOS")

    def test_gos_file(self):
        gos_file_path = pooch.retrieve(
            url=_GOSH_URL,
            known_hash=_GOSH_KNOWN_HASH,
        )
        self.s.create_model(auto_add_edges=True, gos_file_path=gos_file_path)

    def test_auto_add_background_true(self):
        m = self.s.create_model(auto_background=True)
        from hyperspy.components1d import PowerLaw

        is_pl_instance = [isinstance(c, PowerLaw) for c in m]
        assert True in is_pl_instance

    def test_auto_add_edges_false(self):
        m = self.s.create_model(auto_background=False)
        from hyperspy.components1d import PowerLaw

        is_pl_instance = [isinstance(c, PowerLaw) for c in m]
        assert not True in is_pl_instance

    def test_auto_add_edges_false_names(self):
        m = self.s.create_model(auto_add_edges=False)
        cnames = [component.name for component in m]
        assert not "B_K" in cnames or "C_K" in cnames

    def test_convolved_ll_not_set(self):
        m = self.s.create_model(auto_add_edges=False)
        with pytest.raises(RuntimeError, match="not set"):
            m.convolved = True

    def test_low_loss(self):
        s = self.s
        low_loss = s.deepcopy()
        low_loss.axes_manager[-1].offset = -20
        m = s.create_model(low_loss=low_loss)
        assert m.low_loss is low_loss
        assert m.convolved

    def test_low_loss_bad_shape(self):
        low_loss = hs.stack([self.s] * 2)
        with pytest.raises(ValueError):
            _ = self.s.create_model(low_loss=low_loss)


@lazifyTestClass
class TestEELSModel:
    def setup_method(self, method):
        s = EELSSpectrum(np.ones(200))
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        s.add_elements(("B", "C"))
        self.s = s
        self.m = s.create_model()

    def test_suspend_auto_fsw(self):
        m = self.m
        m["B_K"].fine_structure_width = 140.0
        m.suspend_auto_fine_structure_width()
        m.enable_fine_structure()
        m.resolve_fine_structure()
        assert 140 == m["B_K"].fine_structure_width

    def test_resume_fsw(self):
        m = self.m
        m["B_K"].fine_structure_width = 140.0
        m.suspend_auto_fine_structure_width()
        m.resume_auto_fine_structure_width()
        window = (
            m["C_K"].onset_energy.value
            - m["B_K"].onset_energy.value
            - m._preedge_safe_window_width
        )
        m.enable_fine_structure()
        m.resolve_fine_structure()
        assert window == m["B_K"].fine_structure_width

    def test_disable_fine_structure(self):
        self.m.components.C_K.fine_structure_active = True
        self.m.components.B_K.fine_structure_active = True
        self.m.disable_fine_structure()
        assert not self.m.components.C_K.fine_structure_active
        assert not self.m.components.B_K.fine_structure_active

    def test_get_first_ionization_edge_energy_C_B(self):
        assert (
            self.m._get_first_ionization_edge_energy()
            == self.m["B_K"].onset_energy.value
        )

    def test_get_first_ionization_edge_energy_C(self):
        self.m["B_K"].active = False
        assert (
            self.m._get_first_ionization_edge_energy()
            == self.m["C_K"].onset_energy.value
        )

    def test_get_first_ionization_edge_energy_None(self):
        self.m["B_K"].active = False
        self.m["C_K"].active = False
        assert self.m._get_first_ionization_edge_energy() is None

    def test_two_area_powerlaw_estimation_BC(self):
        self.m.signal.data = 2.0 * self.m.axis.axis ** (-3)  # A= 2, r=3
        # self.m.signal.axes_manager[-1].is_binned = False
        self.m.two_area_background_estimation()
        np.testing.assert_allclose(
            self.m._background_components[0].A.value, 2.1451237089380295
        )
        np.testing.assert_allclose(
            self.m._background_components[0].r.value, 3.0118980767392736
        )

    def test_two_area_powerlaw_estimation_C(self):
        self.m["B_K"].active = False
        self.m.signal.data = 2.0 * self.m.axis.axis ** (-3)  # A= 2, r=3
        # self.m.signal.axes_manager[-1].is_binned = False
        self.m.two_area_background_estimation()
        np.testing.assert_allclose(
            self.m._background_components[0].A.value, 2.3978438900878087
        )
        np.testing.assert_allclose(
            self.m._background_components[0].r.value, 3.031884021065014
        )

    def test_two_area_powerlaw_estimation_no_edge(self):
        self.m["B_K"].active = False
        self.m["C_K"].active = False
        self.m.signal.data = 2.0 * self.m.axis.axis ** (-3)  # A= 2, r=3
        print(self.m.signal.axes_manager[-1].is_binned)
        # self.m.signal.axes_manager[-1].is_binned = False
        self.m.two_area_background_estimation()
        np.testing.assert_allclose(
            self.m._background_components[0].A.value, 2.6598803469440986
        )
        np.testing.assert_allclose(
            self.m._background_components[0].r.value, 3.0494030409062058
        )

    def test_get_start_energy_none(self):
        assert self.m._get_start_energy() == 150

    def test_get_start_energy_above(self):
        assert self.m._get_start_energy(170) == 170

    def test_get_start_energy_below(self):
        assert self.m._get_start_energy(100) == 150

    def test_remove_components(self):
        comp = self.m[1]
        assert len(self.m) == 3
        self.m.remove(comp)
        assert len(self.m) == 2

    def test_fit_wrong_kind(self):
        with pytest.raises(ValueError):
            self.m.fit(kind="wrongkind")

    def test_enable_background(self):
        self.m.components.PowerLaw.active = False
        self.m.enable_background()
        assert self.m.components.PowerLaw.active

    def test_disable_background(self):
        self.m.components.PowerLaw.active = True
        self.m.disable_background()
        assert not self.m.components.PowerLaw.active

    def test_signal1d_property(self):
        assert self.s == self.m.signal
        s_new = EELSSpectrum(np.ones(200))
        s_new.set_microscope_parameters(100, 10, 10)
        self.m.signal = s_new
        assert self.m.signal == s_new

    def test_signal1d_property_wrong_value_setter(self):
        m = self.m
        s = hs.signals.Signal1D(np.ones(200))
        with pytest.raises(ValueError):
            self.m.signal = s

    def test_remove(self):
        m = self.m
        c_k = m.components.C_K
        assert c_k in m
        m.remove(c_k)
        assert c_k not in m

    def test_quantify(self):
        f = io.StringIO()
        with contextlib.redirect_stdout(f):
            self.m.quantify()
        out = f.getvalue()
        assert (
            out
            == "\nAbsolute quantification:\nElem.\tIntensity\nB\t1.000000\nC\t1.000000\n"
        )

    def test_enable_edges(self):
        m = self.m
        m.components.B_K.active = False
        m.components.C_K.active = False
        m.enable_edges(edges_list=[m.components.B_K])
        assert m.components.B_K.active
        assert not m.components.C_K.active
        m.enable_edges()
        assert m.components.B_K.active
        assert m.components.C_K.active

    def test_disable_edges(self):
        m = self.m
        m.components.B_K.active = True
        m.components.C_K.active = True
        m.disable_edges(edges_list=[m.components.B_K])
        assert not m.components.B_K.active
        assert m.components.C_K.active
        m.disable_edges()
        assert not m.components.B_K.active
        assert not m.components.C_K.active

    def test_set_all_edges_intensities_positive(self):
        m = self.m
        m.components.B_K.intensity.ext_force_positive = False
        m.components.B_K.intensity.ext_bounded = False
        m.components.C_K.intensity.ext_force_positive = False
        m.components.C_K.intensity.ext_bounded = False
        m.set_all_edges_intensities_positive()
        assert m.components.B_K.intensity.ext_force_positive
        assert m.components.B_K.intensity.ext_bounded
        assert m.components.C_K.intensity.ext_force_positive
        assert m.components.C_K.intensity.ext_bounded

    def test_unset_all_edges_intensities_positive(self):
        m = self.m
        m.components.B_K.intensity.ext_force_positive = True
        m.components.B_K.intensity.ext_bounded = True
        m.components.C_K.intensity.ext_force_positive = True
        m.components.C_K.intensity.ext_bounded = True
        m.unset_all_edges_intensities_positive()
        assert not m.components.B_K.intensity.ext_force_positive
        assert not m.components.B_K.intensity.ext_bounded
        assert not m.components.C_K.intensity.ext_force_positive
        assert not m.components.C_K.intensity.ext_bounded

    def test_fix_edges(self):
        m = self.m
        m.components.B_K.onset_energy.free = True
        m.components.B_K.intensity.free = True
        m.components.B_K.fine_structure_coeff.free = True
        m.components.C_K.onset_energy.free = True
        m.components.C_K.intensity.free = True
        m.components.C_K.fine_structure_coeff.free = True
        m.fix_edges(edges_list=[m.components.B_K])
        assert not m.components.B_K.onset_energy.free
        assert not m.components.B_K.intensity.free
        assert not m.components.B_K.fine_structure_coeff.free
        assert m.components.C_K.onset_energy.free
        assert m.components.C_K.intensity.free
        assert m.components.C_K.fine_structure_coeff.free
        m.fix_edges()
        assert not m.components.B_K.onset_energy.free
        assert not m.components.B_K.intensity.free
        assert not m.components.B_K.fine_structure_coeff.free
        assert not m.components.C_K.onset_energy.free
        assert not m.components.C_K.intensity.free
        assert not m.components.C_K.fine_structure_coeff.free

    def test_free_edges(self):
        m = self.m
        m.components.B_K.onset_energy.free = False
        m.components.B_K.intensity.free = False
        m.components.B_K.fine_structure_coeff.free = False
        m.components.C_K.onset_energy.free = False
        m.components.C_K.intensity.free = False
        m.components.C_K.fine_structure_coeff.free = False
        m.free_edges(edges_list=[m.components.B_K])
        assert m.components.B_K.onset_energy.free
        assert m.components.B_K.intensity.free
        assert m.components.B_K.fine_structure_coeff.free
        assert not m.components.C_K.onset_energy.free
        assert not m.components.C_K.intensity.free
        assert not m.components.C_K.fine_structure_coeff.free
        m.free_edges()
        assert m.components.B_K.onset_energy.free
        assert m.components.B_K.intensity.free
        assert m.components.B_K.fine_structure_coeff.free
        assert m.components.C_K.onset_energy.free
        assert m.components.C_K.intensity.free
        assert m.components.C_K.fine_structure_coeff.free

    def test_fix_fine_structure(self):
        m = self.m
        m.components.B_K.fine_structure_coeff.free = True
        m.components.C_K.fine_structure_coeff.free = True
        m.fix_fine_structure(edges_list=[m.components.B_K])
        assert not m.components.B_K.fine_structure_coeff.free
        assert m.components.C_K.fine_structure_coeff.free
        m.fix_fine_structure()
        assert not m.components.B_K.fine_structure_coeff.free
        assert not m.components.C_K.fine_structure_coeff.free

    def test_free_fine_structure(self):
        m = self.m
        m.components.B_K.fine_structure_coeff.free = False
        m.components.C_K.fine_structure_coeff.free = False
        m.free_fine_structure(edges_list=[m.components.B_K])
        assert m.components.B_K.fine_structure_coeff.free
        assert not m.components.C_K.fine_structure_coeff.free
        m.free_fine_structure()
        assert m.components.B_K.fine_structure_coeff.free
        assert m.components.C_K.fine_structure_coeff.free


@lazifyTestClass
class TestEELSModelFitting:
    def setup_method(self, method):
        data = np.zeros(200)
        data[25:] = 100
        s = EELSSpectrum(data)
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        s.add_elements(("B",))
        self.m = s.create_model(auto_background=False)

    @pytest.mark.parametrize("kind", ["std", "smart"])
    def test_free_edges(self, kind):
        m = self.m
        m.enable_fine_structure()
        intensity = m.components.B_K.intensity.value
        onset_energy = m.components.B_K.onset_energy.value
        fine_structure_coeff = m.components.B_K.fine_structure_coeff.value
        m.free_edges()
        m.multifit(kind=kind)
        assert intensity != m.components.B_K.intensity.value
        assert onset_energy != m.components.B_K.onset_energy.value
        assert fine_structure_coeff != m.components.B_K.fine_structure_coeff.value

    @pytest.mark.parametrize("kind", ["std", "smart"])
    def test_fix_edges(self, kind):
        m = self.m
        m.enable_fine_structure()
        intensity = m.components.B_K.intensity.value
        onset_energy = m.components.B_K.onset_energy.value
        fine_structure_coeff = m.components.B_K.fine_structure_coeff.value
        m.free_edges()
        m.fix_edges()
        m.multifit(kind=kind)
        assert intensity == m.components.B_K.intensity.value
        assert onset_energy == m.components.B_K.onset_energy.value
        assert fine_structure_coeff == m.components.B_K.fine_structure_coeff.value


@lazifyTestClass
class TestFitBackground:
    def setup_method(self, method):
        s = EELSSpectrum(np.ones(200))
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        CE = elements.C.Atomic_properties.Binding_energies.K.onset_energy_eV
        BE = elements.B.Atomic_properties.Binding_energies.K.onset_energy_eV
        s.isig[BE:] += 1
        s.isig[CE:] += 1
        s.add_elements(("Be", "B", "C"))
        self.m = s.create_model(auto_background=False)
        self.m.append(hs.model.components1D.Offset())

    def test_fit_background_B_C(self):
        self.m.fit_background()
        np.testing.assert_allclose(self.m["Offset"].offset.value, 1)
        assert self.m["B_K"].active
        assert self.m["C_K"].active

    def test_fit_background_C(self):
        self.m["B_K"].active = False
        self.m.fit_background()
        np.testing.assert_allclose(self.m["Offset"].offset.value, 1.7142857)
        assert not self.m["B_K"].active
        assert self.m["C_K"].active

    def test_fit_background_no_edge(self):
        self.m["B_K"].active = False
        self.m["C_K"].active = False
        self.m.fit_background()
        np.testing.assert_allclose(self.m["Offset"].offset.value, 2.14)
        assert not self.m["B_K"].active
        assert not self.m["C_K"].active


@lazifyTestClass
class TestFitBackground2D:
    def setup_method(self):
        pl = hs.model.components1D.PowerLaw()
        data = np.empty((2, 250))
        data[0] = pl.function(np.arange(150, 400))
        pl.r.value = 1.5
        data[1] = pl.function(np.arange(150, 400))
        s = EELSSpectrum(data)
        s.set_microscope_parameters(100, 10, 10)
        s.axes_manager[-1].offset = 150
        self.s = s
        self.m = s.create_model()

    def test_only_current_false(self):
        self.m.fit_background(only_current=False)
        residual = self.s - self.m.as_signal()
        assert pytest.approx(residual.data) == 0


@lazifyTestClass
class TestEELSFineStructure:
    def setup_method(self, method):
        s = EELSSpectrum(np.zeros((1024)))
        s.axes_manager[0].units = "eV"
        s.axes_manager[0].scale = 0.1
        s.axes_manager[0].offset = 690
        s.add_elements(["Fe"])
        s.set_microscope_parameters(100, 15, 30)

        m = s.create_model(GOS="hydrogenic", auto_background=False)

        self.g1 = hs.model.components1D.GaussianHF(centre=712, height=50, fwhm=3)
        self.g2 = hs.model.components1D.GaussianHF(centre=725, height=30, fwhm=4)
        self.m = m

    def test_fs_components_in_model_update(self):
        self.m.components.Fe_L3.fine_structure_components.update((self.g1, self.g2))
        for component in self.m.components.Fe_L3.fine_structure_components:
            assert component in self.m

    def test_fs_components_in_model_add(self):
        self.m.components.Fe_L3.fine_structure_components.update((self.g1, self.g2))
        for component in (self.g1, self.g2):
            self.m.components.Fe_L3.fine_structure_components.add(component)
            assert component in self.m

    @pytest.mark.parametrize("fine_structure_active", [True, False])
    def test_fs_components_inherit_fs_active(self, fine_structure_active):
        self.m.components.Fe_L3.fine_structure_active = fine_structure_active
        self.m.components.Fe_L3.fine_structure_components.update((self.g1, self.g2))
        for component in self.m.components.Fe_L3.fine_structure_components:
            assert component.active == fine_structure_active
        self.m.components.Fe_L3.fine_structure_active = not fine_structure_active
        for component in self.m.components.Fe_L3.fine_structure_components:
            assert component.active == (not fine_structure_active)

    def test_fine_structure_smoothing(self):
        Fe = self.m.components.Fe_L3
        Fe.fine_structure_active = True
        Fe.fine_structure_smoothing = 0.3
        len_coeff = len(Fe.fine_structure_coeff.value)
        Fe.fine_structure_smoothing = 0.2
        assert len(Fe.fine_structure_coeff) < len_coeff
        Fe.fine_structure_smoothing = 0.4
        assert len(Fe.fine_structure_coeff) > len_coeff
        with pytest.raises(ValueError):
            Fe.fine_structure_smoothing = 3
        with pytest.raises(ValueError):
            Fe.fine_structure_smoothing = -3

    def test_free_fix_fine_structure(self):
        Fe = self.m.components.Fe_L3
        Fe.fine_structure_active = True
        assert Fe.fine_structure_coeff.free
        Fe.fix_fine_structure()
        assert not Fe.fine_structure_coeff.free
        Fe.free_fine_structure()
        assert Fe.fine_structure_coeff.free
        Fe.fine_structure_components.update((self.g1, self.g2))
        self.g1.fwhm.free = False
        self.g2.fwhm.free = False
        Fe.fix_fine_structure()
        for component in (self.g1, self.g2):
            for parameter in component.parameters:
                assert not parameter.free
        Fe.free_fine_structure()
        for component in (self.g1, self.g2):
            for parameter in component.parameters:
                if parameter.name != "fwhm":
                    assert parameter.free
                else:
                    assert not parameter.free

    def test_fine_structure_active_frees_coeff(self):
        Fe = self.m.components.Fe_L3
        Fe.fine_structure_active = True
        assert Fe.fine_structure_coeff.free
        Fe.fine_structure_coeff.free = False
        Fe.fine_structure_active = False
        assert not Fe.fine_structure_coeff.free
        Fe.fine_structure_active = True
        assert not Fe.fine_structure_coeff.free

    def test_fine_structure_spline_active_frees_coeff(self):
        Fe = self.m.components.Fe_L3
        Fe.fine_structure_active = True
        assert Fe.fine_structure_coeff.free
        Fe.fine_structure_spline_active = False
        assert not Fe.fine_structure_coeff.free
        Fe.fine_structure_spline_active = True
        assert Fe.fine_structure_coeff.free
        Fe.fine_structure_coeff.free = False
        Fe.fine_structure_spline_active = False
        assert not Fe.fine_structure_coeff.free
        Fe.fine_structure_spline_active = True
        assert not Fe.fine_structure_coeff.free

    def test_fine_structure_spline(self):
        v2i = self.m.signal.axes_manager[0].value2index
        Fe = self.m.components.Fe_L3
        Fe.fine_structure_active = True
        Fe.fine_structure_spline_active = True
        Fe.fine_structure_width = 30
        onset = Fe.onset_energy.value
        axis1 = np.linspace(
            Fe.onset_energy.value,
            Fe.onset_energy.value + Fe.fine_structure_width,
            endpoint=False,
        )
        assert np.all(Fe.function(axis1) == 0)
        Fe.fine_structure_coeff.value = (
            np.arange(len(Fe.fine_structure_coeff.value)) + 1
        )
        assert np.all(Fe.function(axis1) != 0)
        Fe.fine_structure_spline_onset = 10
        Fe.fine_structure_coeff.value = (
            np.arange(len(Fe.fine_structure_coeff.value)) + 1
        )
        axis2 = np.linspace(
            Fe.onset_energy.value,
            Fe.onset_energy.value + Fe.fine_structure_spline_onset,
            endpoint=False,
        )
        axis3 = np.linspace(
            Fe.onset_energy.value + Fe.fine_structure_spline_onset,
            Fe.onset_energy.value + Fe.fine_structure_width,
            endpoint=False,
        )
        assert np.all(Fe.function(axis2) == 0)
        assert np.all(Fe.function(axis3) != 0)

    def test_model_store_restore(self):
        Fe = self.m.components.Fe_L3
        Fe.fine_structure_active = True
        Fe.fine_structure_components.update((self.g1, self.g2))
        Fe.fine_structure_spline_onset = 20
        Fe.fine_structure_coeff.value = (
            np.arange(len(Fe.fine_structure_coeff.value)) + 1
        )
        m = self.m
        m.store()
        mc = m.signal.models.a.restore()
        assert np.array_equal(m._get_current_data(), mc._get_current_data())


class TestModelJacobians:
    def setup_method(self, method):
        s = EELSSpectrum(np.zeros(1))
        m = s.create_model(auto_add_edges=False, auto_background=False)
        self.low_loss = 7.0
        self.weights = 0.3
        m.axis.axis = np.array([1, 0])
        m._channel_switches = np.array([0, 1], dtype=bool)
        m.append(hs.model.components1D.Gaussian())
        m[0].A.value = 1
        m[0].centre.value = 2.0
        m[0].sigma.twin = m[0].centre
        m._low_loss = mock.MagicMock()
        m._low_loss._get_current_data = mock.MagicMock()
        m.low_loss._get_current_data.return_value = self.low_loss
        self.model = m
        m.convolution_axis = np.zeros(2)

    def test_jacobian_convolved(self):
        m = self.model
        m.convolved = True
        m.append(hs.model.components1D.Gaussian())
        m[0].convolved = False
        m[1].convolved = True
        assert m.low_loss._get_current_data() == 7
        jac = m._jacobian((1, 2, 3, 4, 5), None, weights=self.weights)
        np.testing.assert_array_almost_equal(
            jac.squeeze(),
            self.weights
            * np.array(
                [
                    m[0].A.grad(0),
                    m[0].sigma.grad(0) + m[0].centre.grad(0),
                    m[1].A.grad(0) * self.low_loss,
                    m[1].centre.grad(0) * self.low_loss,
                    m[1].sigma.grad(0) * self.low_loss,
                ]
            ),
        )
        assert m[0].A.value == 1
        assert m[0].centre.value == 2
        assert m[0].sigma.value == 2
        assert m[1].A.value == 3
        assert m[1].centre.value == 4
        assert m[1].sigma.value == 5


class TestModelSettingPZero:
    def setup_method(self, method):
        s = EELSSpectrum(np.empty(1))
        m = s.create_model(auto_add_edges=False, auto_background=False)
        self.model = m

    def test_calculating_convolution_axis(self):
        m = self.model
        # setup
        m.axis.offset = 10
        m.axis.size = 10
        ll_axis = mock.MagicMock()
        ll_axis.size = 7
        ll_axis.value2index.return_value = 3
        m._low_loss = mock.MagicMock()
        m.low_loss.axes_manager.signal_axes = [
            ll_axis,
        ]

        # calculation
        m.set_convolution_axis()

        # tests
        np.testing.assert_array_equal(m.convolution_axis, np.arange(7, 23))
        np.testing.assert_equal(ll_axis.value2index.call_args[0][0], 0)


@lazifyTestClass
class TestConvolveModelSlicing:
    def setup_method(self, method):
        s = EELSSpectrum(np.random.random((10, 10, 600)))
        s.axes_manager[-1].offset = -150.0
        s.axes_manager[-1].scale = 0.5
        m = s.create_model(auto_add_edges=False, auto_background=False)
        m.low_loss = s + 1
        g = hs.model.components1D.Gaussian()
        m.append(g)
        self.m = m

    def test_slicing_low_loss_inav(self):
        m = self.m
        m1 = m.inav[::2]
        assert m1.signal.data.shape == m1.low_loss.data.shape

    def test_slicing_low_loss_isig(self):
        m = self.m
        m1 = m.isig[::2]
        assert m.signal.data.shape == m1.low_loss.data.shape


class TestModelDictionary:
    def setup_method(self, method):
        s = EELSSpectrum(np.array([1.0, 2, 4, 7, 12, 7, 4, 2, 1]))
        m = s.create_model(auto_add_edges=False, auto_background=False)
        m.low_loss = (s + 3.0).deepcopy()
        self.model = m
        self.s = s

        m.append(hs.model.components1D.Gaussian())
        m.append(hs.model.components1D.Gaussian())
        m.append(hs.model.components1D.ScalableFixedPattern(s * 0.3))
        m[0].A.twin = m[1].A
        m.fit()

    def test_to_dictionary(self):
        m = self.model
        d = m.as_dictionary()

        np.testing.assert_allclose(m.low_loss.data, d["low_loss"]["data"])

    def test_load_dictionary(self):
        d = self.model.as_dictionary()
        mn = self.s.create_model()
        mn.append(hs.model.components1D.Lorentzian())
        mn._load_dictionary(d)
        mo = self.model

        np.testing.assert_allclose(mn.low_loss.data, mo.low_loss.data)
        for i in range(len(mn)):
            assert mn[i]._id_name == mo[i]._id_name
            for po, pn in zip(mo[i].parameters, mn[i].parameters):
                np.testing.assert_allclose(po.map["values"], pn.map["values"])
                np.testing.assert_allclose(po.map["is_set"], pn.map["is_set"])

        assert mn[0].A.twin is mn[1].A



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import math
import logging

import numpy as np
from scipy import integrate, interpolate, constants

from exspy.misc.eels.base_gos import BaseGOS

_logger = logging.getLogger(__name__)

XU = [
    0.82,
    0.52,
    0.52,
    0.42,
    0.30,
    0.29,
    0.22,
    0.30,
    0.22,
    0.16,
    0.12,
    0.13,
    0.13,
    0.14,
    0.16,
    0.18,
    0.19,
    0.22,
    0.14,
    0.11,
    0.12,
    0.12,
    0.12,
    0.10,
    0.10,
    0.10,
]
# IE3=[73,99,135,164,200,245,294,347,402,455,513,575,641,710,
# 779,855,931,1021,1115,1217,1323,1436,1550,1675]

# IE1=[118,149,189,229,270,320,377,438,500,564,628,695,769,846,
# 926,1008,1096,1194,1142,1248,1359,1476,1596,1727]

R = constants.value("Rydberg constant times hc in eV")


class HydrogenicGOS(BaseGOS):

    """Computes the K and L GOS using R. Egerton's  routines.

    Parameters
    ----------
    element_subshell : str
        For example, 'Ti_L3' for the GOS of the titanium L3 subshell

    Methods
    -------
    parametrize_GOS()
        Parametrize the GOS to speed up the calculation.
    get_qaxis_and_gos(ienergy, qmin, qmax)
        Given the energy axis index and qmin and qmax values returns
        the qaxis and gos between qmin and qmax using linear
        interpolation to include qmin and qmax in the range.


    Attributes
    ----------
    energy_axis : array
        The tabulated energy axis
    qaxis : array
        The tabulated qaxis
    energy_onset: float
        The energy onset for the given element subshell as obtained
        from iternal tables.

    Notes
    -----
    The Hydrogeninc GOS are calculated using R. Egerton's SIGMAK3 and
    SIGMAL3 routines that has been translated from Matlab to Python by
    I. Iyengar. See http://www.tem-eels.ca/ for the original code.

    """

    _name = "hydrogenic"

    def __init__(self, element_subshell):
        """
        Parameters
        ----------

        element_subshell : str
            For example, 'Ti_L3' for the GOS of the titanium L3 subshell

        """
        # Check if the Peter Rez's Hartree Slater GOS distributed by
        # Gatan are available. Otherwise exit

        self.element, self.subshell = element_subshell.split("_")
        self.read_elements()
        self.energy_shift = 0

        if self.subshell[:1] == "K":
            self.gosfunc = self.gosfuncK
            self.rel_energy_axis = self.get_parametrized_energy_axis(50, 3, 50)
        elif self.subshell[:1] == "L":
            self.gosfunc = self.gosfuncL
            self.onset_energy_L3 = self.element_dict["Atomic_properties"][
                "Binding_energies"
            ]["L3"]["onset_energy (eV)"]
            self.onset_energy_L1 = self.element_dict["Atomic_properties"][
                "Binding_energies"
            ]["L1"]["onset_energy (eV)"]
            self.onset_energy = self.onset_energy_L3
            relative_axis = self.get_parametrized_energy_axis(50, 3, 50)
            dL3L2 = self.onset_energy_L1 - self.onset_energy_L3
            self.rel_energy_axis = np.hstack(
                (
                    relative_axis[: relative_axis.searchsorted(dL3L2)],
                    relative_axis + dL3L2,
                )
            )
        else:
            raise ValueError(
                "The Hydrogenic GOS currently can only"
                "compute K or L shells. Try using other GOS."
            )

        self.energy_axis = self.rel_energy_axis + self.onset_energy

        info_str = (
            "\nHydrogenic GOS\n"
            + ("\tElement: %s " % self.element)
            + ("\tSubshell: %s " % self.subshell)
            + ("\tOnset Energy = %s " % self.onset_energy)
        )
        _logger.info(info_str)

    def integrateq(self, onset_energy, angle, E0):
        energy_shift = onset_energy - self.onset_energy
        self.energy_shift = energy_shift
        gamma = 1 + E0 / 511.06
        T = 511060 * (1 - 1 / gamma**2) / 2
        qint = np.zeros((self.energy_axis.shape[0]))
        for i, E in enumerate(self.energy_axis + energy_shift):
            qa0sqmin = (E**2) / (4 * R * T) + (E**3) / (8 * gamma**3 * R * T**2)
            p02 = T / (R * (1 - 2 * T / 511060))
            pp2 = p02 - E / R * (gamma - E / 1022120)
            qa0sqmax = qa0sqmin + 4 * np.sqrt(p02 * pp2) * (math.sin(angle / 2)) ** 2

            # dsbyde IS THE ENERGY-DIFFERENTIAL X-SECN (barn/eV/atom)
            qint[i] = (
                3.5166e8
                * (R / T)
                * (R / E)
                * (
                    integrate.quad(
                        lambda x: self.gosfunc(E, np.exp(x)),
                        math.log(qa0sqmin),
                        math.log(qa0sqmax),
                    )[0]
                )
            )
        self.qint = qint
        return interpolate.make_interp_spline(
            self.energy_axis + energy_shift,
            qint,
            k=1,
        )

    def gosfuncK(self, E, qa02):
        # gosfunc calculates (=DF/DE) which IS PER EV AND PER ATOM
        z = self.Z
        r = 13.606
        zs = 1.0
        rnk = 1
        if z != 1:
            zs = z - 0.5
            rnk = 2

        q = qa02 / zs**2
        kh2 = E / (r * zs**2) - 1
        akh = np.sqrt(abs(kh2))
        if akh < 0.01:
            akh = 0.01
        if kh2 >= 0.0:
            d = 1 - np.e ** (-2 * np.pi / akh)
            bp = np.arctan(2 * akh / (q - kh2 + 1))
            if bp < 0:
                bp = bp + np.pi
            c = np.e ** ((-2 / akh) * bp)
        else:
            d = 1
            y = -1 / akh * np.log((q + 1 - kh2 + 2 * akh) / (q + 1 - kh2 - 2 * akh))
            c = np.e**y
        a = ((q - kh2 + 1) ** 2 + 4 * kh2) ** 3
        return 128 * rnk * E / (r * zs**4) * c / d * (q + kh2 / 3 + 1 / 3) / (a * r)

    def gosfuncL(self, E, qa02):
        # gosfunc calculates (=DF/DE) which IS PER EV AND PER ATOM
        # Note: quad function only works with qa02 due to IF statements in
        # function

        z = self.Z
        r = 13.606
        zs = z - 0.35 * (8 - 1) - 1.7
        iz = z - 11
        if iz >= len(XU):
            # Egerton does not tabulate the correction for Z>36.
            # This produces XSs that are within 10% of Hartree-Slater XSs
            # for these elements.
            u = 0.1
        else:
            # Egerton's correction to the Hydrogenic XS
            u = XU[int(iz)]
        el3 = self.onset_energy_L3 + self.energy_shift
        el1 = self.onset_energy_L1 + self.energy_shift

        q = qa02 / zs**2
        kh2 = E / (r * zs**2) - 0.25
        akh = np.sqrt(abs(kh2))
        if kh2 >= 0.0:
            d = 1 - np.exp(-2 * np.pi / akh)
            bp = np.arctan(akh / (q - kh2 + 0.25))
            if bp < 0:
                bp = bp + np.pi
            c = np.exp((-2 / akh) * bp)
        else:
            d = 1
            y = -1 / akh * np.log((q + 0.25 - kh2 + akh) / (q + 0.25 - kh2 - akh))
            c = np.exp(y)

        if E - el1 <= 0:
            g = (
                2.25 * q**4
                - (0.75 + 3 * kh2) * q**3
                + (0.59375 - 0.75 * kh2 - 0.5 * kh2**2) * q * q
                + (0.11146 + 0.85417 * kh2 + 1.8833 * kh2 * kh2 + kh2**3) * q
                + 0.0035807
                + kh2 / 21.333
                + kh2 * kh2 / 4.5714
                + kh2**3 / 2.4
                + kh2**4 / 4
            )

            a = ((q - kh2 + 0.25) ** 2 + kh2) ** 5
        else:
            g = (
                q**3
                - (5 / 3 * kh2 + 11 / 12) * q**2
                + (kh2 * kh2 / 3 + 1.5 * kh2 + 65 / 48) * q
                + kh2**3 / 3
                + 0.75 * kh2 * kh2
                + 23 / 48 * kh2
                + 5 / 64
            )
            a = ((q - kh2 + 0.25) ** 2 + kh2) ** 4
        rf = ((E + 0.1 - el3) / 1.8 / z / z) ** u
        # The following commented lines are to give a more accurate GOS
        # for edges presenting white lines. However, this is not relevant
        # for quantification by curve fitting.
        # if abs(iz - 11) <= 5 and E - el3 <= 20:
        #     rf = 1
        return rf * 32 * g * c / a / d * E / r / r / zs**4



"""
Modules containing the exSpy signals and their lazy counterparts.

EELSSpectrum
    For electron energy-loss data with ``signal_dimension`` equal one, i.e.
    spectral data of ``n`` dimensions. The signal is binned by default.
EDSTEMSpectrum
    For electron energy-dispersive X-ray data acquired in a transmission
    electron microscopy with ``signal_dimension`` equal one, i.e.
    spectral data of ``n`` dimensions. The signal is binned by default.
EDSSEMSpectrum
    For electron energy-dispersive X-ray data acquired in a scanning
    electron microscope with ``signal_dimension`` equal one, i.e.
    spectral data of ``n`` dimensions. The signal is binned by default.
DielectricFunction
    For dielectric function data with ``signal_dimension`` equal one. The signal
    is unbinned by default.
"""
from .dielectric_function import DielectricFunction, LazyDielectricFunction
from .eds import EDSSpectrum, LazyEDSSpectrum
from .eds_sem import EDSSEMSpectrum, LazyEDSSEMSpectrum
from .eds_tem import EDSTEMSpectrum, LazyEDSTEMSpectrum
from .eels import EELSSpectrum, LazyEELSSpectrum


__all__ = [
    "DielectricFunction",
    "LazyDielectricFunction",
    "EDSSpectrum",
    "LazyEDSSpectrum",
    "EDSTEMSpectrum",
    "LazyEDSTEMSpectrum",
    "EELSSpectrum",
    "LazyEELSSpectrum",
    "EDSSEMSpectrum",
    "LazyEDSSEMSpectrum",
]


def __dir__():
    return sorted(__all__)



import logging
import pytest

from hyperspy.utils.markers import Lines, Texts
from hyperspy.misc.test_utils import update_close_figure
from hyperspy.utils import stack

import exspy

default_tol = 2.0
baseline_dir = "plot_spectra_markers"
style_pytest_mpl = "default"


class TestEDSMarkers:
    @pytest.mark.mpl_image_compare(
        baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl
    )
    def test_plot_eds_lines(
        self,
    ):
        a = exspy.data.EDS_SEM_TM002()
        s = stack([a, a * 5])
        s.plot(True)
        s.axes_manager.navigation_axes[0].index = 1
        return s._plot.signal_plot.figure

    @pytest.mark.parametrize("norm", [None, "log", "auto", "linear"])
    def test_plot_eds_lines_norm(self, norm):
        a = exspy.data.EDS_SEM_TM002()
        s = stack([a, a * 5])
        # When norm is None, don't specify (use default)
        # otherwise use specify value
        kwargs = {"norm": norm} if norm else {}
        s.plot(True, **kwargs)

    @pytest.mark.mpl_image_compare(
        baseline_dir=baseline_dir,
        tolerance=default_tol,
        style=style_pytest_mpl,
        filename="test_plot_eds_lines.png",
    )
    def test_plot_xray_lines(
        self,
    ):
        # It should be the same image as with previous test (test_plot_eds_lines)
        a = exspy.data.EDS_SEM_TM002()
        s = stack([a, a * 5])
        s.plot()
        s._plot_xray_lines(xray_lines=True)
        s.axes_manager.navigation_axes[0].index = 1
        return s._plot.signal_plot.figure

    def test_plot_eds_lines_not_in_range(self, caplog):
        s = exspy.data.EDS_SEM_TM002().isig[5.0:8.0]
        s.plot()
        with caplog.at_level(logging.WARNING):
            s._plot_xray_lines(xray_lines=["Pt_Ka"])

        assert "Pt_Ka is not in the data energy range." in caplog.text

    def test_plot_eds_lines_background(
        self,
    ):
        s = exspy.data.EDS_SEM_TM002().isig[5.0:8.0]
        s.plot()
        bw = s.estimate_background_windows()
        s._plot_xray_lines(background_windows=bw)

    def test_plot_add_background_windows(
        self,
    ):
        s = exspy.data.EDS_SEM_TM002().isig[5.0:8.0]
        s.plot()
        bw = s.estimate_background_windows()
        s._add_background_windows_markers(bw)
        # Add integration windows
        iw = s.estimate_integration_windows(windows_width=2.0, xray_lines=["Fe_Ka"])
        s._add_vertical_lines_groups(iw, linestyle="--")

    def test_plot_eds_markers_no_energy(
        self,
    ):
        s = exspy.data.EDS_SEM_TM002()
        del s.metadata.Acquisition_instrument.SEM.beam_energy
        s.plot(True)


class TestEELSMarkers:
    @pytest.mark.mpl_image_compare(
        baseline_dir=baseline_dir, tolerance=default_tol, style=style_pytest_mpl
    )
    def test_plot_eels_labels(
        self,
    ):
        s = exspy.data.EELS_MnFe(True, add_noise=False)
        s.add_elements(["Cr"])
        s.plot(plot_edges=True)
        return s._plot.signal_plot.figure

    def test_plot_eels_labels_nav(
        self,
    ):
        s = exspy.data.EELS_MnFe(True, add_noise=False)
        s.add_elements(["Cr", "Fe"])
        s.plot(plot_edges=True)
        s.axes_manager.indices = (9,)
        s._plot.close()

    def test_remove_edges(self):
        s = exspy.data.EELS_MnFe(True, add_noise=False)
        s.plot(plot_edges=["Cr"])

        lines = s._edge_markers["lines"]
        texts = s._edge_markers["texts"]

        assert isinstance(lines, Lines)
        assert isinstance(texts, Texts)
        assert len(lines) == 8
        assert len(texts) == 8

        s._remove_edge_labels()
        assert len(lines) == 0
        assert len(texts) == 0


@update_close_figure()
def test_plot_eds_markers_close():
    s = exspy.data.EDS_SEM_TM002()
    s.plot(True)
    return s



import numpy as np
import traits.api as t

from hyperspy.exceptions import SignalDimensionError
from hyperspy.ui_registry import add_gui_method
from hyperspy.signal_tools import SpanSelectorInSignal1D
from exspy.misc.eels.tools import get_edges_near_energy, get_info_from_edges


@add_gui_method(toolkey="exspy.EELSSpectrum.print_edges_table")
class EdgesRange(SpanSelectorInSignal1D):
    units = t.Unicode()
    edges_list = t.Tuple()
    only_major = t.Bool()
    order = t.Unicode("closest")
    complementary = t.Bool(True)

    def __init__(self, signal, interactive=True):
        if signal.axes_manager.signal_dimension != 1:
            raise SignalDimensionError(signal.axes_manager.signal_dimension, 1)

        if interactive:
            super().__init__(signal)
        else:
            # ins non-interactive mode, don't initialise the span selector
            self.signal = signal
            self.axis = self.signal.axes_manager.signal_axes[0]

        self.active_edges = list(self.signal._edge_markers["names"])
        self.active_complementary_edges = []
        self.units = self.axis.units
        self.btns = []

        if self.signal._edge_markers["lines"] is None:
            self.signal._initialise_markers()

        self._get_edges_info_within_energy_axis()

        self.signal.axes_manager.events.indices_changed.connect(
            self._on_navigation_indices_changed, []
        )
        self.signal._plot.signal_plot.events.closed.connect(
            lambda: self.signal.axes_manager.events.indices_changed.disconnect(
                self._on_navigation_indices_changed
            ),
            [],
        )

    def _get_edges_info_within_energy_axis(self):
        mid_energy = (self.axis.low_value + self.axis.high_value) / 2
        rng = self.axis.high_value - self.axis.low_value
        self.edge_all = np.asarray(
            get_edges_near_energy(mid_energy, rng, order=self.order)
        )
        info = get_info_from_edges(self.edge_all)

        energy_all = []
        relevance_all = []
        description_all = []
        for d in info:
            onset = d["onset_energy (eV)"]
            relevance = d["relevance"]
            threshold = d["threshold"]
            edge_ = d["edge"]
            description = threshold + ". " * (threshold != "" and edge_ != "") + edge_

            energy_all.append(onset)
            relevance_all.append(relevance)
            description_all.append(description)

        self.energy_all = np.asarray(energy_all)
        self.relevance_all = np.asarray(relevance_all)
        self.description_all = np.asarray(description_all)

    def _on_navigation_indices_changed(self):
        self.signal._plot.signal_plot.update()

    def update_table(self):
        if self.span_selector is not None:
            energy_mask = (self.ss_left_value <= self.energy_all) & (
                self.energy_all <= self.ss_right_value
            )
            if self.only_major:
                relevance_mask = self.relevance_all == "Major"
            else:
                relevance_mask = np.ones(len(self.edge_all), bool)

            mask = energy_mask & relevance_mask
            self.edges_list = tuple(self.edge_all[mask])
            energy = tuple(self.energy_all[mask])
            relevance = tuple(self.relevance_all[mask])
            description = tuple(self.description_all[mask])
        else:
            self.edges_list = ()
            energy, relevance, description = (), (), ()

        self._keep_valid_edges()

        return self.edges_list, energy, relevance, description

    def _keep_valid_edges(self):
        edge_all = list(self.signal._edge_markers["names"])
        for edge in edge_all:
            if edge not in self.edges_list:
                if edge in self.active_edges:
                    self.active_edges.remove(edge)
                elif edge in self.active_complementary_edges:
                    self.active_complementary_edges.remove(edge)
                self.signal._remove_edge_labels([edge], render_figure=False)
            elif edge not in self.active_edges:
                self.active_edges.append(edge)

        self._on_complementary()
        self._update_labels()

    def update_active_edge(self, change):
        state = change["new"]
        edge = change["owner"].description

        if state:
            self.active_edges.append(edge)
        else:
            if edge in self.active_edges:
                self.active_edges.remove(edge)
            if edge in self.active_complementary_edges:
                self.active_complementary_edges.remove(edge)
            self.signal._remove_edge_labels([edge], render_figure=False)
        self._on_complementary()
        self._update_labels()

    def _on_complementary(self):
        if self.complementary:
            self.active_complementary_edges = self.signal._get_complementary_edges(
                self.active_edges, self.only_major
            )
        else:
            self.active_complementary_edges = []

    def check_btn_state(self):
        edges = [btn.description for btn in self.btns]
        for btn in self.btns:
            edge = btn.description
            if btn.value is False:
                if edge in self.active_edges:
                    self.active_edges.remove(edge)
                    self.signal._remove_edge_labels([edge])
                if edge in self.active_complementary_edges:
                    btn.value = True

            if btn.value is True and self.complementary:
                comp = self.signal._get_complementary_edges(
                    self.active_edges, self.only_major
                )
                for cedge in comp:
                    if cedge in edges:
                        pos = edges.index(cedge)
                        self.btns[pos].value = True

    def _update_labels(self, active=None, complementary=None):
        # update selected and/or complementary edges
        if active is None:
            active = self.active_edges
        if complementary is None:
            complementary = self.active_complementary_edges

        edges_on_signal = set(self.signal._edge_markers["names"])
        edges_to_show = set(set(active).union(complementary))
        edge_keep = edges_on_signal.intersection(edges_to_show)
        edge_remove = edges_on_signal.difference(edge_keep)
        edge_add = edges_to_show.difference(edge_keep)

        if edge_remove:
            # Remove edges out
            self.signal._remove_edge_labels(edge_remove, render_figure=False)
        if edge_add:
            # Add the new edges
            self.signal._add_edge_labels(edge_add, render_figure=False)
        if edge_remove or edge_add:
            # Render figure only once
            self.signal._render_figure(plot=["signal_plot"])

    def _clear_markers(self):
        # Used in hyperspy_gui_ipywidgets
        self.signal._remove_edge_labels()
        self.active_edges = []
        self.active_complementary_edges = []



"""Components """

from .eels_arctan import EELSArctan
from .eels_cl_edge import EELSCLEdge
from .eels_double_power_law import DoublePowerLaw
from .eels_vignetting import Vignetting
from .pes_core_line_shape import PESCoreLineShape
from .pes_see import SEE
from .pes_voigt import PESVoigt
from .volume_plasmon_drude import VolumePlasmonDrude


__all__ = [
    "EELSArctan",
    "EELSCLEdge",
    "DoublePowerLaw",
    "PESCoreLineShape",
    "PESVoigt",
    "SEE",
    "Vignetting",
    "VolumePlasmonDrude",
]


def __dir__():
    return sorted(__all__)



import numpy as np
import pytest

import hyperspy.api as hs
from hyperspy import _lazy_signals
from hyperspy.decorators import lazifyTestClass

import exspy


class TestConvertSigna:
    def setup_method(self, method):
        self.s = hs.signals.Signal1D([0, 1])

    def test_lazy_to_eels_and_back(self):
        self.s = self.s.as_lazy()
        self.s.set_signal_type("EELS")
        assert isinstance(self.s, exspy.signals.LazyEELSSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, _lazy_signals.LazySignal1D)

    def test_signal1d_to_eels(self):
        self.s.set_signal_type("EELS")
        assert isinstance(self.s, exspy.signals.EELSSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.Signal1D)

    def test_signal1d_to_eds_tem(self):
        self.s.set_signal_type("EDS_TEM")
        assert isinstance(self.s, exspy.signals.EDSTEMSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.Signal1D)

    def test_signal1d_to_eds_sem(self):
        self.s.set_signal_type("EDS_SEM")
        assert isinstance(self.s, exspy.signals.EDSSEMSpectrum)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.Signal1D)

    def test_error_None(self):
        with pytest.raises(TypeError):
            self.s.set_signal_type(None)


class TestConvertComplexSignal1D:
    def setup_method(self, method):
        self.s = hs.signals.ComplexSignal1D([0, 1])

    def test_complex_to_dielectric_function(self):
        self.s.set_signal_type("DielectricFunction")
        assert isinstance(self.s, exspy.signals.DielectricFunction)
        self.s.set_signal_type("")
        assert isinstance(self.s, hs.signals.ComplexSignal1D)


def test_lazy_to_eels_and_back():
    s = hs.signals.Signal1D([0, 1])
    s = s.as_lazy()
    s.set_signal_type("EELS")
    assert isinstance(s, exspy.signals.LazyEELSSpectrum)
    s.set_signal_type("")
    assert isinstance(s, _lazy_signals.LazySignal1D)


def test_complex_to_dielectric_function():
    s = hs.signals.ComplexSignal1D([0, 1])
    s.set_signal_type("DielectricFunction")
    assert isinstance(s, exspy.signals.DielectricFunction)
    s.set_signal_type("")
    assert isinstance(s, hs.signals.ComplexSignal1D)


@lazifyTestClass
class Test1d:
    def setup_method(self, method):
        self.s = hs.signals.BaseSignal([0, 1, 2])

    def test_set_EELS(self):
        s = self.s.as_signal1D(0)
        s.set_signal_type("EELS")
        assert s.metadata.Signal.signal_type == "EELS"
        if s._lazy:
            _class = exspy.signals.LazyEELSSpectrum
        else:
            _class = exspy.signals.EELSSpectrum
        assert isinstance(s, _class)


@lazifyTestClass
class Test2d:
    def setup_method(self, method):
        self.s = hs.signals.BaseSignal(np.random.random((2, 3)))  # (|3, 2)

    def test_s2EELS2im2s(self):
        pytest.importorskip("exspy")
        s = self.s.as_signal1D(0)
        s.set_signal_type("EELS")
        im = s.as_signal2D((1, 0))
        assert im.metadata.Signal.signal_type == "EELS"
        s = im.as_signal1D(0)
        assert s.metadata.Signal.signal_type == "EELS"
        if s._lazy:
            from exspy.signals import LazyEELSSpectrum

            _class = LazyEELSSpectrum
        else:
            from exspy.signals import EELSSpectrum

            _class = EELSSpectrum
        assert isinstance(s, _class)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest

import hyperspy.api as hs

from exspy.components import PESCoreLineShape


def test_PESCoreLineShape():
    core_line = PESCoreLineShape(A=10, FWHM=1.5, origin=0.5)
    x = np.linspace(-5, 15, 10)
    np.testing.assert_allclose(
        core_line.function(x),
        np.array(
            [
                8.97054744e-04,
                0.365234208,
                7.09463858,
                6.57499512,
                0.290714653,
                6.13260141e-04,
                6.17204216e-08,
                2.96359844e-13,
                6.78916184e-20,
                7.42026292e-28,
            ]
        ),
    )
    assert core_line._position is core_line.origin


def test_PESCoreLineShape_shirley():
    core_line = PESCoreLineShape(A=10, FWHM=1.5, origin=0.5)
    core_line.Shirley = True
    core_line.shirley.value = 0.01
    x = np.linspace(-5, 15, 10)
    np.testing.assert_allclose(
        core_line.function(x),
        np.array(
            [
                0.144159014,
                0.504843825,
                7.16330182,
                6.57790840,
                0.290720786,
                6.13260758e-04,
                6.17204245e-08,
                2.96359844e-13,
                6.78916184e-20,
                7.42026292e-28,
            ]
        ),
    )
    np.testing.assert_allclose(core_line.function(x), core_line.function_nd(x))


@pytest.mark.parametrize("Shirley", [False, True])
def test_PESCoreLineShape_fit(Shirley):
    # Component parameter values
    A = 10
    FWHM = 0.5
    origin = 5.0
    shirley = 0.01 if Shirley else 0.0

    offset, scale, size = 0, 0.1, 100
    x = np.linspace(offset, scale * size, size)
    comp = PESCoreLineShape(A=A, FWHM=FWHM, origin=origin)
    comp.Shirley = Shirley
    comp.shirley.value = shirley

    s = hs.signals.Signal1D(comp.function(x))
    axis = s.axes_manager[0]
    axis.offset, axis.scale = offset, scale
    s.add_gaussian_noise(0.1, random_state=1)
    m = s.create_model()
    core_line = PESCoreLineShape(A=1, FWHM=1.5, origin=0.5)
    core_line.Shirley = Shirley
    m.append(core_line)
    m.fit(grad="analytical")
    np.testing.assert_allclose(core_line.A.value, A, rtol=0.1)
    np.testing.assert_allclose(abs(core_line.FWHM.value), FWHM, rtol=0.1)
    np.testing.assert_allclose(core_line.origin.value, origin, rtol=0.1)
    np.testing.assert_allclose(core_line.shirley.value, shirley, rtol=0.1)


@pytest.mark.parametrize("Shirley", [False, True])
def test_PESCoreLineShape_function_nd(Shirley):
    A, FWHM, origin = 10, 1.5, 0.0
    core_line = PESCoreLineShape(A=A, FWHM=FWHM, origin=origin)
    core_line.Shirley = Shirley
    core_line.shirley.value = 0.01 if Shirley else 0.0
    x = np.linspace(-5, 15, 1000)
    s = hs.signals.Signal1D(np.array([x] * 2))

    # Manually set to test function_nd
    core_line._axes_manager = s.axes_manager
    core_line._create_arrays()
    core_line.A.map["values"] = [A] * 2
    core_line.FWHM.map["values"] = [FWHM] * 2
    core_line.origin.map["values"] = [origin] * 2
    core_line.shirley.map["values"] = [core_line.shirley.value] * 2

    values = core_line.function_nd(x)
    assert values.shape == (2, len(x))
    for v in values:
        np.testing.assert_allclose(v, core_line.function(x), rtol=0.5)


@pytest.mark.parametrize("Shirley", [False, True])
def test_recreate_component(Shirley):
    core_line = PESCoreLineShape(A=10, FWHM=1.5, origin=0.5)
    core_line.Shirley = Shirley

    s = hs.signals.Signal1D(np.zeros(10))
    m = s.create_model()
    m.append(core_line)
    model_dict = m.as_dictionary()

    m2 = s.create_model()
    m2._load_dictionary(model_dict)
    assert m2[0].Shirley == Shirley



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging
from pathlib import Path

import numpy as np
from scipy import constants

from hyperspy.defaults_parser import preferences
from exspy.misc.eels.base_gos import TabulatedGOS


_logger = logging.getLogger(__name__)

R = constants.value("Rydberg constant times hc in eV")
a0 = constants.value("Bohr radius")

# This dictionary accounts for conventions chosen in naming the data files, as well as normalisation.
# These cross sections contain only odd-number edges such as N3, or M5, and are normalised accordingly.
# Other edges can be obtained as scaled copies of the provided ones.
conventions = {
    "K": {"table": "K1", "factor": 1},
    "L1": {"table": "L1", "factor": 1},
    "L2,3": {"table": "L3", "factor": 3 / 2},
    "L2": {"table": "L3", "factor": 1 / 2},
    "L3": {"table": "L3", "factor": 1},
    "M1": {"table": "M1", "factor": 1},
    "M2,3": {"table": "M3", "factor": 3 / 2},
    "M2": {"table": "M3", "factor": 1 / 2},
    "M3": {"table": "M3", "factor": 1},
    "M4,5": {"table": "M5", "factor": 5 / 3},
    "M4": {"table": "M5", "factor": 2 / 3},
    "M5": {"table": "M5", "factor": 1},
    "N1": {"table": "N1", "factor": 1},
    "N2,3": {"table": "N3", "factor": 3 / 2},
    "N2": {"table": "N3", "factor": 1 / 2},
    "N3": {"table": "N3", "factor": 1},
    "N4,5": {"table": "N5", "factor": 5 / 3},
    "N4": {"table": "N5", "factor": 2 / 3},
    "N5": {"table": "N5", "factor": 1},
    "N6,7": {"table": "N7", "factor": 7 / 4},
    "N6": {"table": "N7", "factor": 4 / 7},
    "N7": {"table": "N7", "factor": 1},
    "O1": {"table": "O1", "factor": 1},
    "O2,3": {"table": "O3", "factor": 3 / 2},
    "O2": {"table": "O3", "factor": 1 / 2},
    "O3": {"table": "O3", "factor": 1},
    "O4,5": {"table": "O5", "factor": 5 / 3},
    "O4": {"table": "O5", "factor": 2 / 3},
    "O5": {"table": "O5", "factor": 1},
}


class HartreeSlaterGOS(TabulatedGOS):

    """Read Hartree-Slater Generalized Oscillator Strength parametrized
    from files.

    Parameters
    ----------
    element_subshell : {str, dict}
        Usually a string, for example, 'Ti_L3' for the GOS of the titanium L3
        subshell. If a dictionary is passed, it is assumed that Hartree Slater
        GOS was exported using `GOS.as_dictionary`, and will be reconstructed.

    Methods
    -------
    readgosfile()
        Read the GOS files of the element subshell from the location
        defined in Preferences.
    get_qaxis_and_gos(ienergy, qmin, qmax)
        given the energy axis index and qmin and qmax values returns
        the qaxis and gos between qmin and qmax using linear
        interpolation to include qmin and qmax in the range.

    Attributes
    ----------
    energy_axis : array
        The tabulated energy axis
    qaxis : array
        The tabulated qaxis
    energy_onset: float
        The energy onset for the given element subshell as obtained
        from iternal tables.

    """

    _name = "Hartree-Slater"
    _whitelist = {
        "gos_array": None,
        "rel_energy_axis": None,
        "qaxis": None,
        "element": None,
        "subshell": None,
    }

    def read_elements(self):
        super().read_elements()
        self.subshell_factor = conventions[self.subshell]["factor"]

    def read_gos_data(self):  # pragma: no cover
        _logger.info(
            "Hartree-Slater GOS\n"
            f"\tElement: {self.element} "
            f"\tSubshell: {self.subshell}"
            f"\tOnset Energy = {self.onset_energy}"
        )
        element = self.element
        subshell = self.subshell
        table = conventions[subshell]["table"]

        # Check if the Peter Rez's Hartree Slater GOS distributed by
        # Gatan are available. Otherwise exit
        gos_root = Path(preferences.EELS.eels_gos_files_path)
        gos_file = gos_root / f"{element}.{table}"

        if not gos_root.is_dir():
            raise FileNotFoundError(
                "Parametrized Hartree-Slater GOS files not "
                f"found in {gos_root}. Please define a valid "
                "location for the files in the preferences as "
                "`preferences.EELS.eels_gos_files_path`."
            )

        with open(gos_file) as f:
            GOS_list = f.read().replace("\r", "").split()

        # Map the parameters
        info1_1 = float(GOS_list[2])
        info1_2 = float(GOS_list[3])
        ncol = int(GOS_list[5])
        info2_1 = float(GOS_list[6])
        info2_2 = float(GOS_list[7])
        nrow = int(GOS_list[8])
        self.gos_array = np.array(GOS_list[9:], dtype=float)
        # The division by R is not in the equations, but it seems that
        # the the GOS was tabulated this way
        self.gos_array = self.gos_array.reshape(nrow, ncol) / R
        del GOS_list

        # Calculate the scale of the matrix
        self.rel_energy_axis = self.get_parametrized_energy_axis(info2_1, info2_2, nrow)
        self.qaxis = self.get_parametrized_qaxis(info1_1, info1_2, ncol)
        self.energy_axis = self.rel_energy_axis + self.onset_energy



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import math
import numbers
import logging

import numpy as np
import matplotlib.pyplot as plt
from scipy import constants

from hyperspy.misc.array_tools import rebin
from exspy.misc.elements import elements as elements_db
import hyperspy.defaults_parser

_logger = logging.getLogger(__name__)


def _estimate_gain(
    ns, cs, weighted=False, higher_than=None, plot_results=False, binning=0, pol_order=1
):
    if binning > 0:
        factor = 2**binning
        remainder = np.mod(ns.shape[1], factor)
        if remainder != 0:
            ns = ns[:, remainder:]
            cs = cs[:, remainder:]
        new_shape = (ns.shape[0], ns.shape[1] / factor)
        ns = rebin(ns, new_shape)
        cs = rebin(cs, new_shape)

    noise = ns - cs
    variance = np.var(noise, 0)
    average = np.mean(cs, 0).squeeze()

    # Select only the values higher_than for the calculation
    if higher_than is not None:
        sorting_index_array = np.argsort(average)
        average_sorted = average[sorting_index_array]
        average_higher_than = average_sorted > higher_than
        variance_sorted = variance.squeeze()[sorting_index_array]
        variance2fit = variance_sorted[average_higher_than]
        average2fit = average_sorted[average_higher_than]
    else:
        variance2fit = variance
        average2fit = average

    fit = np.polyfit(average2fit, variance2fit, pol_order)
    if weighted is True:
        from hyperspy._signals.signal1D import Signal1D
        from hyperspy.models.model1d import Model1D
        from hyperspy.components1d import Line

        s = Signal1D(variance2fit)
        s.axes_manager.signal_axes[0].axis = average2fit
        m = Model1D(s)
        l = Line()
        l.a.value = fit[1]
        l.b.value = fit[0]
        m.append(l)
        m.fit(weights=True)
        fit[0] = l.b.value
        fit[1] = l.a.value

    if plot_results is True:
        plt.figure()
        plt.scatter(average.squeeze(), variance.squeeze())
        plt.xlabel("Counts")
        plt.ylabel("Variance")
        plt.plot(average2fit, np.polyval(fit, average2fit), color="red")
    results = {"fit": fit, "variance": variance.squeeze(), "counts": average.squeeze()}

    return results


def _estimate_correlation_factor(g0, gk, k):
    a = math.sqrt(g0 / gk)
    e = k * (a - 1) / (a - k)
    c = (1 - e) ** 2
    return c


def estimate_variance_parameters(
    noisy_signal,
    clean_signal,
    mask=None,
    pol_order=1,
    higher_than=None,
    return_results=False,
    plot_results=True,
    weighted=False,
    store_results="ask",
):
    """Find the scale and offset of the Poissonian noise

    By comparing an SI with its denoised version (i.e. by PCA),
    this plots an
    estimation of the variance as a function of the number of counts
    and fits a
    polynomy to the result.

    Parameters
    ----------
    noisy_SI, clean_SI : signal1D.Signal1D instances
    mask : numpy bool array
        To define the channels that will be used in the calculation.
    pol_order : int
        The order of the polynomy.
    higher_than: float
        To restrict the fit to counts over the given value.
    return_results : Bool
    plot_results : Bool
    store_results: {True, False, "ask"}, default "ask"
        If True, it stores the result in the signal metadata

    Returns
    -------
    Dictionary with the result of a linear fit to estimate the offset
    and scale factor

    """
    with noisy_signal.unfolded(), clean_signal.unfolded():
        # The rest of the code assumes that the first data axis
        # is the navigation axis. We transpose the data if that is not the
        # case.
        ns = (
            noisy_signal.data.copy()
            if noisy_signal.axes_manager[0].index_in_array == 0
            else noisy_signal.data.T.copy()
        )
        cs = (
            clean_signal.data.copy()
            if clean_signal.axes_manager[0].index_in_array == 0
            else clean_signal.data.T.copy()
        )

        if mask is not None:
            _slice = [
                slice(None),
            ] * len(ns.shape)
            _slice[noisy_signal.axes_manager.signal_axes[0].index_in_array] = ~mask
            ns = ns[_slice]
            cs = cs[_slice]

        results0 = _estimate_gain(
            ns,
            cs,
            weighted=weighted,
            higher_than=higher_than,
            plot_results=plot_results,
            binning=0,
            pol_order=pol_order,
        )

        results2 = _estimate_gain(
            ns,
            cs,
            weighted=weighted,
            higher_than=higher_than,
            plot_results=False,
            binning=2,
            pol_order=pol_order,
        )

        c = _estimate_correlation_factor(results0["fit"][0], results2["fit"][0], 4)

        message = (
            "Gain factor: %.2f\n" % results0["fit"][0]
            + "Gain offset: %.2f\n" % results0["fit"][1]
            + "Correlation factor: %.2f\n" % c
        )
        if store_results == "ask":
            is_ok = ""
            while is_ok not in ("Yes", "No"):
                is_ok = input(message + "Would you like to store the results (Yes/No)?")
            is_ok = is_ok == "Yes"
        else:
            is_ok = store_results
            _logger.info(message)
        if is_ok:
            noisy_signal.metadata.set_item(
                "Signal.Noise_properties.Variance_linear_model.gain_factor",
                results0["fit"][0],
            )
            noisy_signal.metadata.set_item(
                "Signal.Noise_properties.Variance_linear_model.gain_offset",
                results0["fit"][1],
            )
            noisy_signal.metadata.set_item(
                "Signal.Noise_properties.Variance_linear_model." "correlation_factor", c
            )
            noisy_signal.metadata.set_item(
                "Signal.Noise_properties.Variance_linear_model."
                + "parameters_estimation_method",
                "exSpy",
            )

    if return_results is True:
        return results0


def power_law_perc_area(E1, E2, r):
    a = E1
    b = E2
    return (
        100
        * (
            (a**r * r - a**r)
            * (a / (a**r * r - a**r) - (b + a) / ((b + a) ** r * r - (b + a) ** r))
        )
        / a
    )


def rel_std_of_fraction(a, std_a, b, std_b, corr_factor=1):
    rel_a = std_a / a
    rel_b = std_b / b
    return np.sqrt(rel_a**2 + rel_b**2 - 2 * rel_a * rel_b * corr_factor)


def ratio(edge_A, edge_B):
    a = edge_A.intensity.value
    std_a = edge_A.intensity.std
    b = edge_B.intensity.value
    std_b = edge_B.intensity.std
    ratio = a / b
    ratio_std = ratio * rel_std_of_fraction(a, std_a, b, std_b)
    _logger.info(
        "Ratio %s/%s %1.3f +- %1.3f ", edge_A.name, edge_B.name, a / b, 1.96 * ratio_std
    )
    return ratio, ratio_std


def eels_constant(s, zlp, t):
    r"""Calculate the constant of proportionality (k) in the relationship
    between the EELS signal and the dielectric function.
    dielectric function from a single scattering distribution (SSD) using
    the Kramers-Kronig relations.

        .. math::

            S(E)=\frac{I_{0}t}{\pi a_{0}m_{0}v^{2}}\ln\left[1+\left(\frac{\beta}
            {\theta_{E}}\right)^{2}\right]\Im(\frac{-1}{\epsilon(E)})=
            k\Im(\frac{-1}{\epsilon(E)})


    Parameters
    ----------
    zlp: {number, BaseSignal}
        If the ZLP is the same for all spectra, the intengral of the ZLP
        can be provided as a number. Otherwise, if the ZLP intensity is not
        the same for all spectra, it can be provided as i) a Signal
        of the same dimensions as the current signal containing the ZLP
        spectra for each location ii) a Signal of signal dimension 0
        and navigation_dimension equal to the current signal containing the
        integrated ZLP intensity.
    t: {None, number, BaseSignal}
        The sample thickness in nm. If the thickness is the same for all
        spectra it can be given by a number. Otherwise, it can be provided
        as a Signal with signal dimension 0 and navigation_dimension equal
        to the current signal.

    Returns
    -------
    k: Signal instance

    """

    # Constants and units
    me = constants.value("electron mass energy equivalent in MeV") * 1e3  # keV

    # Mapped parameters
    try:
        e0 = s.metadata.Acquisition_instrument.TEM.beam_energy
    except BaseException:
        raise AttributeError(
            "Please define the beam energy."
            "You can do this e.g. by using the "
            "set_microscope_parameters method"
        )
    try:
        beta = s.metadata.Acquisition_instrument.TEM.Detector.EELS.collection_angle
    except BaseException:
        raise AttributeError(
            "Please define the collection semi-angle."
            "You can do this e.g. by using the "
            "set_microscope_parameters method"
        )

    axis = s.axes_manager.signal_axes[0]
    eaxis = axis.axis.copy()
    if eaxis[0] == 0:
        # Avoid singularity at E=0
        eaxis[0] = 1e-10

    if isinstance(zlp, hyperspy.signal.BaseSignal):
        if zlp.axes_manager.navigation_dimension == s.axes_manager.navigation_dimension:
            if zlp.axes_manager.signal_dimension == 0:
                i0 = zlp.data
            else:
                i0 = zlp.integrate1D(axis.index_in_axes_manager).data
        else:
            raise ValueError(
                "The ZLP signal dimensions are not "
                "compatible with the dimensions of the "
                "low-loss signal"
            )
        # The following prevents errors if the signal is a single spectrum
        if len(i0) != 1:
            i0 = i0.reshape(np.insert(i0.shape, axis.index_in_array, 1))
    elif isinstance(zlp, numbers.Number):
        i0 = zlp
    else:
        raise ValueError(
            "The zero-loss peak input is not valid, it must be\
                         in the BaseSignal class or a Number."
        )

    if isinstance(t, hyperspy.signal.BaseSignal):
        if (
            t.axes_manager.navigation_dimension == s.axes_manager.navigation_dimension
        ) and (t.axes_manager.signal_dimension == 0):
            t = t.data
            t = t.reshape(np.insert(t.shape, axis.index_in_array, 1))
        else:
            raise ValueError(
                "The thickness signal dimensions are not "
                "compatible with the dimensions of the "
                "low-loss signal"
            )

    # Kinetic definitions
    ke = e0 * (1 + e0 / 2.0 / me) / (1 + e0 / me) ** 2
    tgt = e0 * (2 * me + e0) / (me + e0)
    k = s.__class__(
        data=(t * i0 / (332.5 * ke)) * np.log(1 + (beta * tgt / eaxis) ** 2)
    )
    k.metadata.General.title = "EELS proportionality constant K"
    return k


def get_edges_near_energy(energy, width=10, only_major=False, order="closest"):
    """Find edges near a given energy that are within the given energy
    window.

    Parameters
    ----------
    energy : float
        Energy to search, in eV
    width : float
        Width of window, in eV, around energy in which to find nearby
        energies, i.e. a value of 10 eV (the default) means to
        search +/- 5 eV. The default is 10.
    only_major : bool
        Whether to show only the major edges. The default is False.
    order : str
        Sort the edges, if 'closest', return in the order of energy difference,
        if 'ascending', return in ascending order, similarly for 'descending'

    Returns
    -------
    edges : list
        All edges that are within the given energy window, sorted by
        energy difference to the given energy.
    """

    if width < 0:
        raise ValueError("Provided width needs to be >= 0.")
    if order not in ("closest", "ascending", "descending"):
        raise ValueError("order needs to be 'closest', 'ascending' or " "'descending'")

    Emin, Emax = energy - width / 2, energy + width / 2

    # find all subshells that have its energy within range
    valid_edges = []
    for element, element_info in elements_db.items():
        try:
            for shell, shell_info in element_info["Atomic_properties"][
                "Binding_energies"
            ].items():
                if only_major:
                    if shell_info["relevance"] != "Major":
                        continue
                if shell[-1] != "a" and Emin <= shell_info["onset_energy (eV)"] <= Emax:
                    subshell = "{}_{}".format(element, shell)
                    Ediff = abs(shell_info["onset_energy (eV)"] - energy)
                    valid_edges.append(
                        (subshell, shell_info["onset_energy (eV)"], Ediff)
                    )
        except KeyError:
            continue

    # Sort according to 'order' and return only the edges
    if order == "closest":
        edges = [edge for edge, _, _ in sorted(valid_edges, key=lambda x: x[2])]
    elif order == "ascending":
        edges = [edge for edge, _, _ in sorted(valid_edges, key=lambda x: x[1])]
    elif order == "descending":
        edges = [
            edge for edge, _, _ in sorted(valid_edges, key=lambda x: x[1], reverse=True)
        ]

    return edges


def get_info_from_edges(edges):
    """Return the information of a sequence of edges as a list of dictionaries

    Parameters
    ----------
    edges : str or iterable
        the sequence of edges, each entry in the format of 'element_subshell'.

    Returns
    -------
    info : list
        a list of dictionaries with information corresponding to the provided
        edges.
    """

    edges = np.atleast_1d(edges)
    info = []
    for edge in edges:
        element, subshell = edge.split("_")
        d = elements_db[element]["Atomic_properties"]["Binding_energies"][subshell]
        info.append(d)

    return info



#!/usr/bin/env python3
import argparse
import re
import subprocess


def run_towncrier(tag):
    cmd = ("towncrier", "build", "--version", tag.strip("v"))

    return subprocess.call(cmd)


def update_fallback_version_in_pyproject(tag, fname="pyproject.toml"):
    version = tag.strip("v").split(".")
    # Default to +1 on minor version
    major, minor = version[0], int(version[1]) + 1

    with open(fname, "r") as file:
        lines = file.readlines()

    pattern = "fallback_version"
    new_version = f"{major}.{minor}.dev0"
    # Iterate through the lines and find the pattern
    for i, line in enumerate(lines):
        if re.search(pattern, line):
            lines[i] = f'{pattern} = "{new_version}"\n'
            break

    # Write the updated content back to the file
    with open(fname, "w") as file:
        file.writelines(lines)

    print(
        f"\nNew (fallback) dev version ({new_version}) written to `pyproject.toml`.\n"
    )


if __name__ == "__main__":
    # Get tag argument
    parser = argparse.ArgumentParser()
    parser.add_argument("tag")
    args = parser.parse_args()
    tag = args.tag

    # Update release notes
    run_towncrier(tag)

    # Update fallback version for setuptools_scm
    update_fallback_version_in_pyproject(tag)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import logging

import h5py
import numpy as np
import pooch
from scipy import constants

from hyperspy.defaults_parser import preferences
from exspy.misc.eels.base_gos import TabulatedGOS


_logger = logging.getLogger(__name__)

R = constants.value("Rydberg constant times hc in eV")
a0 = constants.value("Bohr radius")

_GOSH_DOI = "10.5281/zenodo.7645765"
_GOSH_URL = f"doi:{_GOSH_DOI}/Segger_Guzzinati_Kohl_1.5.0.gosh"
_GOSH_KNOWN_HASH = "md5:7fee8891c147a4f769668403b54c529b"


class GoshGOS(TabulatedGOS):
    """Read Generalized Oscillator Strength from a GOSH database.

    Parameters
    ----------
    element_subshell : {str, dict}
        Usually a string, for example, 'Ti_L3' for the GOS of the titanium L3
        subshell. If a dictionary is passed, it is assumed that a GOSH GOS was
        exported using `GOS.as_dictionary`, and will be reconstructed.

    Methods
    -------
    readgosarray()
        Read the GOS files of the element subshell from the location
        defined in Preferences.
    get_qaxis_and_gos(ienergy, qmin, qmax)
        given the energy axis index and qmin and qmax values returns
        the qaxis and gos between qmin and qmax using linear
        interpolation to include qmin and qmax in the range.

    Attributes
    ----------
    energy_axis : array
        The tabulated energy axis
    qaxis : array
        The tabulated qaxis
    energy_onset: float
        The energy onset for the given element subshell as obtained
        from iternal tables.

    """

    _name = "gosh"
    _whitelist = {
        "gos_array": None,
        "rel_energy_axis": None,
        "qaxis": None,
        "element": None,
        "subshell": None,
    }

    def __init__(self, element_subshell, gos_file_path=None):
        """
        Parameters
        ----------
        element_subshell : str
            For example, 'Ti_L3' for the GOS of the titanium L3 subshell
        gos_file_path : str
            The path of the gosh file to use.
        """

        if gos_file_path is None:
            gos_file_path = pooch.retrieve(
                url=_GOSH_URL,
                known_hash=_GOSH_KNOWN_HASH,
                progressbar=preferences.General.show_progressbar,
            )
        self.gos_file_path = gos_file_path
        super().__init__(element_subshell=element_subshell)

    def read_gos_data(self):
        _logger.info(
            "GOSH precomputed GOS\n"
            f"\tElement: {self.element} "
            f"\tSubshell: {self.subshell}"
            f"\tOnset Energy = {self.onset_energy}"
        )
        element = self.element
        subshell = self.subshell

        error_message = (
            "The GOSH Parametrized GOS database does not "
            f"contain a valid entry the {subshell} subshell "
            f"of {element}. Please select a different database."
        )

        with h5py.File(self.gos_file_path, "r") as h:
            conventions = h["metadata/edges_info"]
            if subshell not in conventions:
                raise ValueError(error_message)
            table = conventions[subshell].attrs["table"]
            self.subshell_factor = conventions[subshell].attrs["occupancy_ratio"]
            stem = f"/{element}/{table}"
            if stem not in h:
                raise ValueError(error_message)
            gos_group = h[stem]
            gos = gos_group["data"][:]
            q = gos_group["q"][:]
            free_energies = gos_group["free_energies"][:]
            doi = h["/metadata/data_ref"].attrs["data_doi"]

        gos = np.squeeze(gos.T)
        self.doi = doi
        self.gos_array = gos
        self.qaxis = q
        self.rel_energy_axis = free_energies - min(free_energies)
        self.energy_axis = self.rel_energy_axis + self.onset_energy



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import pytest

from exspy.misc.eds.utils import _get_element_and_line


def test_get_element_and_line():
    assert _get_element_and_line("Mn_Ka") == ("Mn", "Ka")

    with pytest.raises(ValueError):
        _get_element_and_line("MnKa") == -1



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import math

import numpy as np
from scipy import constants, integrate, interpolate

from exspy.misc.elements import elements
from hyperspy.misc.export_dictionary import (
    export_to_dictionary,
    load_from_dictionary,
)
from hyperspy.misc.math_tools import get_linear_interpolation


R = constants.value("Rydberg constant times hc in eV")
a0 = constants.value("Bohr radius")


class BaseGOS:
    def read_elements(self):
        element = self.element
        subshell = self.subshell
        # Convert to the "GATAN" nomenclature
        if (element in elements) is not True:
            raise ValueError(f"The given element {element} is not in the database.")
        elif subshell not in elements[element]["Atomic_properties"]["Binding_energies"]:
            subshells = ", ".join(
                list(elements[element]["Atomic_properties"]["Binding_energies"].keys())
            )
            raise ValueError(
                f"The given subshell {subshell} is not in the database. The "
                f"available subshells are:\n{subshells}"
            )

        self.onset_energy = elements[element]["Atomic_properties"]["Binding_energies"][
            subshell
        ]["onset_energy (eV)"]
        self.Z = elements[element]["General_properties"]["Z"]
        self.element_dict = elements[element]

    def get_parametrized_qaxis(self, k1, k2, n):
        return k1 * (np.exp(np.arange(n) * k2) - 1) * 1e10

    def get_parametrized_energy_axis(self, k1, k2, n):
        return k1 * (np.exp(np.arange(n) * k2 / k1) - 1)

    def get_qaxis_and_gos(self, ienergy, qmin, qmax):
        qgosi = self.gos_array[ienergy, :]
        if qmax > self.qaxis[-1]:
            # Linear extrapolation
            g1, g2 = qgosi[-2:]
            q1, q2 = self.qaxis[-2:]
            gosqmax = get_linear_interpolation((q1, g1), (q2, g2), qmax)
            qaxis = np.hstack((self.qaxis, qmax))
            qgosi = np.hstack((qgosi, gosqmax))
        else:
            index = self.qaxis.searchsorted(qmax)
            g1, g2 = qgosi[index - 1 : index + 1]
            q1, q2 = self.qaxis[index - 1 : index + 1]
            gosqmax = get_linear_interpolation((q1, g1), (q2, g2), qmax)
            qaxis = np.hstack((self.qaxis[:index], qmax))
            qgosi = np.hstack((qgosi[:index,], gosqmax))

        if qmin > 0:
            index = self.qaxis.searchsorted(qmin)
            g1, g2 = qgosi[index - 1 : index + 1]
            q1, q2 = qaxis[index - 1 : index + 1]
            gosqmin = get_linear_interpolation((q1, g1), (q2, g2), qmin)
            qaxis = np.hstack((qmin, qaxis[index:]))
            qgosi = np.hstack(
                (
                    gosqmin,
                    qgosi[index:],
                )
            )
        return qaxis, qgosi.clip(0)


class TabulatedGOS(BaseGOS):
    def __init__(self, element_subshell):
        """
        Parameters
        ----------
        element_subshell : str
            For example, 'Ti_L3' for the GOS of the titanium L3 subshell

        """
        self.subshell_factor = 1.0
        if isinstance(element_subshell, dict):
            self.element = element_subshell["element"]
            self.subshell = element_subshell["subshell"]
            self.read_elements()
            self._load_dictionary(element_subshell)
        else:
            self.element, self.subshell = element_subshell.split("_")
            self.read_elements()
            self.read_gos_data()

    def _load_dictionary(self, dictionary):
        load_from_dictionary(self, dictionary)
        self.energy_axis = self.rel_energy_axis + self.onset_energy

    def as_dictionary(self, fullcopy=True):
        """Export the GOS as a dictionary."""
        dic = {}
        export_to_dictionary(self, self._whitelist, dic, fullcopy)
        return dic

    def integrateq(self, onset_energy, angle, E0):
        energy_shift = onset_energy - self.onset_energy
        self.energy_shift = energy_shift
        qint = np.zeros((self.energy_axis.shape[0]))
        # Calculate the cross section at each energy position of the
        # tabulated GOS
        gamma = 1 + E0 / 511.06
        T = 511060 * (1 - 1 / gamma**2) / 2
        for i in range(0, self.gos_array.shape[0]):
            E = self.energy_axis[i] + energy_shift
            # Calculate the limits of the q integral
            qa0sqmin = (E**2) / (4 * R * T) + (E**3) / (8 * gamma**3 * R * T**2)
            p02 = T / (R * (1 - 2 * T / 511060))
            pp2 = p02 - E / R * (gamma - E / 1022120)
            qa0sqmax = qa0sqmin + 4 * np.sqrt(p02 * pp2) * (math.sin(angle / 2)) ** 2
            qmin = math.sqrt(qa0sqmin) / a0
            qmax = math.sqrt(qa0sqmax) / a0
            # Perform the integration in a log grid
            qaxis, gos = self.get_qaxis_and_gos(i, qmin, qmax)
            logsqa0qaxis = np.log((a0 * qaxis) ** 2)
            qint[i] = integrate.simps(gos, logsqa0qaxis)
        E = self.energy_axis + energy_shift
        # Energy differential cross section in (barn/eV/atom)
        qint *= (4.0 * np.pi * a0**2.0 * R**2 / E / T * self.subshell_factor) * 1e28
        self.qint = qint
        return interpolate.make_interp_spline(E, qint, k=3)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest

import hyperspy.api as hs

from exspy.components import DoublePowerLaw


def test_function():
    g = DoublePowerLaw()
    g.A.value = 3
    g.r.value = 2
    g.origin.value = 1
    g.shift.value = 2
    g.ratio.value = 2
    assert np.isinf(g.function(1))
    assert np.isinf(g.function(3))
    assert g.function(-1) == 0
    assert g.function(0) == 0
    assert g.function(2) == 9
    np.testing.assert_allclose(g.function(10), 0.15948602)
    assert g.grad_A(2) == 3
    np.testing.assert_allclose(g.grad_r(4), -0.3662041)
    assert g.grad_origin(2) == -6
    assert g.grad_shift(2) == -12
    assert g.grad_ratio(2) == 3


class TestDoublePowerLaw:
    def setup_method(self, method):
        s = hs.signals.Signal1D(np.zeros(1024))
        s.axes_manager[0].offset = 100
        s.axes_manager[0].scale = 0.1
        m = s.create_model()
        exspy = pytest.importorskip("exspy")
        from exspy.components import DoublePowerLaw

        m.append(DoublePowerLaw())
        m[0].A.value = 1000
        m[0].r.value = 4
        m[0].ratio.value = 200
        self.m = m

    @pytest.mark.parametrize(("binned"), (True, False))
    def test_fit(self, binned):
        self.m.signal.axes_manager[-1].is_binned = binned
        s = self.m.as_signal()
        assert s.axes_manager[-1].is_binned == binned
        exspy = pytest.importorskip("exspy")
        g = exspy.components.DoublePowerLaw()
        # Fix the ratio parameter to test the fit
        g.ratio.free = False
        g.shift.free = False
        g.origin.free = False
        g.ratio.value = 200
        m = s.create_model()
        m.append(g)
        m.fit_component(g, signal_range=(None, None))
        np.testing.assert_allclose(g.A.value, 1000.0)
        np.testing.assert_allclose(g.r.value, 4.0)
        np.testing.assert_allclose(g.ratio.value, 200.0)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from exspy.models.edsmodel import EDSModel


class EDSTEMModel(EDSModel):

    """Build and fit a model to EDS data acquired in the TEM.

    Parameters
    ----------
    spectrum : EDSTEMSpectrum

    auto_add_lines : bool
        If True, automatically add Gaussians for all X-rays generated
        in the energy range by an element, using the edsmodel.add_family_lines
        method.
    auto_background : bool
        If True, adds automatically a polynomial order 6 to the model,
        using the edsmodel.add_polynomial_background method.

    Any extra arguments are passed to the Model constructor.
    """

    def __init__(
        self, spectrum, auto_background=True, auto_add_lines=True, *args, **kwargs
    ):
        EDSModel.__init__(
            self, spectrum, auto_background, auto_add_lines, *args, **kwargs
        )
